import numpy as np
# Sampler class is written by referring to https://github.com/pytorch/pytorch/blob/master/torch/utils/data/sampler.py.
class Sampler(object):
    r"""Base class for all Samplers.

    Every Sampler subclass has to provide an :meth:`__iter__` method, providing a
    way to iterate over indices of dataset elements, and a :meth:`__len__` method
    that returns the length of the returned iterators.
    """

    def __init__(self, data_source):
        pass

    def __iter__(self):
        raise NotImplementedError

    def __len__(self):
        return len(self.data_source)

class SequentialSampler(Sampler):
    r"""Samples elements sequentially, always in the same order.

    Arguments:
        data_source (Dataset): dataset to sample from
    """

    def __init__(self, data_source):
        self.data_source = data_source

    def __iter__(self):
        return iter(range(len(self.data_source)))

    def __len__(self) -> int:
        return len(self.data_source)


class RandomSampler(Sampler):
    r"""Samples elements randomly. If without replacement, then sample from a shuffled dataset.
    If with replacement, then user can specify :attr:`num_samples` to draw.

    Arguments:
        data_source (Dataset): dataset to sample from
        replacement (bool): samples are drawn on-demand with replacement if ``True``, default=``False``
        num_samples (int): number of samples to draw, default=`len(dataset)`. This argument
            is supposed to be specified only when `replacement` is ``True``.
    """

    def __init__(self, data_source, replacement=False,
                 num_samples=None):
        self.data_source = data_source
        self.replacement = replacement
        self._num_samples = num_samples

        if not isinstance(self.replacement, bool):
            raise TypeError(">> replacement should be a boolean value, but got "
                            "replacement={}".format(self.replacement))

        if self._num_samples is not None and not replacement:
            raise ValueError(">> With replacement=False, num_samples should not be specified, "
                             "since a random permute will be performed.")

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:
            raise ValueError(">> num_samples should be a positive integer "
                             "value, but got num_samples={}".format(self.num_samples))

    @property
    def num_samples(self):
        # dataset size might change at runtime
        if self._num_samples is None:
            return len(self.data_source)
        return self._num_samples

    def __iter__(self):
        n = len(self.data_source)
        if self.replacement:
            for _ in range(self.num_samples // 32):
                yield from np.random.randint(high=n, size=(32,)).tolist()
            yield from np.random.randint(high=n, size=(self.num_samples % 32,)).tolist()
        else:
            yield from np.random.permutation(np.arange(n)).tolist()

    def __len__(self):
        return self.num_samples

class BatchSampler(Sampler):
    r"""Wraps another sampler to yield a mini-batch of indices.

    Args:
        sampler (Sampler or Iterable): Base sampler. Can be any iterable object
        batch_size (int): Size of mini-batch.
        drop_last (bool): If ``True``, the sampler will drop the last batch if
            its size would be less than ``batch_size``

    Example:
        >>> list(BatchSampler(SequentialSampler(range(10)), batch_size=3, drop_last=False))
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
        >>> list(BatchSampler(SequentialSampler(range(10)), batch_size=3, drop_last=True))
        [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    """

    def __init__(self, sampler, batch_size, drop_last):
        # Since collections.abc.Iterable does not check for `__getitem__`, which
        # is one way for an object to be an iterable, we don't do an `isinstance`
        # check here.
        if not isinstance(sampler, Sampler):
            raise ValueError("sampler should be an instance of "
                             "torch.utils.data.Sampler, but got sampler={}"
                             .format(sampler))
        if not isinstance(batch_size, int) or isinstance(batch_size, bool) or \
                batch_size <= 0:
            raise ValueError("batch_size should be a positive integer value, "
                             "but got batch_size={}".format(batch_size))
        if not isinstance(drop_last, bool):
            raise ValueError("drop_last should be a boolean value, but got "
                             "drop_last={}".format(drop_last))
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        batch = []
        for idx in self.sampler:
            batch.append(idx)
            if len(batch) == self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch

    def __len__(self):
        if self.drop_last:
            return len(self.sampler) // self.batch_size  # type: ignore
        else:
            return (len(self.sampler) + self.batch_size - 1) // self.batch_size  # type: ignore



# -*- coding: utf-8 -*-
"""
Created on 2021/3/30
@project: SPAIC
@filename: STBP_Learner
@author: Mengxiao Zhang
@contact: mxzhangice@gmail.com

@description:
"""
from .Learner import Learner


class STBP(Learner):
    '''
        STBP learning rule.

        Args:
            alpha(num) : The parameter alpha of STBP learning model.
            trainable : The parameter whether it can be trained.

        Methods:
            build(self, backend): Build the backend, realize the algorithm of STBP model.
            threshold(self, x, v_th): Get the threshold of the STBP model.

        Example:
            Net._learner = STBP(0.5, Net)

        Reference:
            Yujie Wu et al. "Spatio-Temporal Backpropagation for Training High-performance Spiking Neural Networks" In:
            Frontiers in Neuroscience, 2018. Volume 12. pp. 331.
            doi:10.3389/fnins.2018.00331
            url:ttps://www.frontiersin.org/article/10.3389/fnins.2018.00331

    '''

    def __init__(self,trainable=None, **kwargs):
        super(STBP, self).__init__(trainable=trainable, **kwargs)
        self.alpha = kwargs.get('alpha', 0.5)
        self.prefered_backend = ['pytorch']
        self.name = 'STBP'
        self.firing_func = None
        self.parameters = kwargs

    def build(self, backend):
        '''
            Build the backend, realize the algorithm of STBP model.
            Args：
                backend: The backend we used to compute.
        '''
        super(STBP, self).build(backend)
        self.device = backend.device0
        if backend.backend_name == 'pytorch':
            import torch
            import math
            class ActFun(torch.autograd.Function):
                """
                Approximate firing func.
                """
                @staticmethod
                def forward(
                        ctx,
                        input,
                        thresh,
                        alpha
                ):
                    import torch
                    import math
                    ctx.thresh = thresh
                    ctx.alpha = alpha
                    ctx.save_for_backward(input)
                    return input.gt(thresh).type_as(input)

                @staticmethod
                def backward(
                        ctx,
                        grad_output
                ):
                    input, = ctx.saved_tensors
                    grad_input = grad_output.clone()
                    ctx.alpha = ctx.alpha.to(input)
                    temp = torch.exp(-(input - ctx.thresh) ** 2 / (2 * ctx.alpha)) \
                           / (2 * math.pi * ctx.alpha)
                    result = grad_input * temp.type_as(input)
                    return result, None, None


            self.firing_func = ActFun()
            self.alpha = torch.tensor(self.alpha).to(self.device)
            # self.backend.basic_operate['threshold'] = self.threshold

            # replace threshold operation in all trainable neuron_groups
            for neuron in self.trainable_groups.values():
                for key in neuron._operations.keys():
                    if 'threshold' in key:
                        neuron._operations[key].func = self.threshold


    def threshold(self, x, v_th):
        '''
            Get the threshold of the STBP model.

            return:
                A method that use STBP model to compute the threshold.

        '''
        return self.firing_func.apply(x, v_th, self.alpha)


Learner.register("stbp", STBP)


# Copyright 2014 Jason Heeris, jason.heeris@gmail.com
# 
# This file is part of the gammatone toolkit, and is licensed under the 3-clause
# BSD license: https://github.com/detly/gammatone/blob/master/COPYING
"""
This module contains functions for constructing sets of equivalent rectangular
bandwidth gammatone filters.
"""
from __future__ import division
from collections import namedtuple

import numpy as np
import scipy as sp
from scipy import signal as sgn

DEFAULT_FILTER_NUM = 100
DEFAULT_LOW_FREQ = 100
DEFAULT_HIGH_FREQ = 44100 / 4


def erb_point(low_freq, high_freq, fraction):
    """
    Calculates a single point on an ERB scale between ``low_freq`` and
    ``high_freq``, determined by ``fraction``. When ``fraction`` is ``1``,
    ``low_freq`` will be returned. When ``fraction`` is ``0``, ``high_freq``
    will be returned.
    
    ``fraction`` can actually be outside the range ``[0, 1]``, which in general
    isn't very meaningful, but might be useful when ``fraction`` is rounded a
    little above or below ``[0, 1]`` (eg. for plot axis labels).
    """
    # Change the following three parameters if you wish to use a different ERB
    # scale. Must change in MakeERBCoeffs too.
    # TODO: Factor these parameters out
    ear_q = 9.26449 # Glasberg and Moore Parameters
    min_bw = 24.7
    order = 1

    # All of the following expressions are derived in Apple TR #35, "An
    # Efficient Implementation of the Patterson-Holdsworth Cochlear Filter
    # Bank." See pages 33-34.
    erb_point = (
        -ear_q * min_bw
        + np.exp(
            fraction * (
                -np.log(high_freq + ear_q * min_bw)
                + np.log(low_freq + ear_q * min_bw)
                )
        ) *
        (high_freq + ear_q * min_bw)
    )
    
    return erb_point


def erb_space(
    low_freq=DEFAULT_LOW_FREQ,
    high_freq=DEFAULT_HIGH_FREQ,
    num=DEFAULT_FILTER_NUM):
    """
    This function computes an array of ``num`` frequencies uniformly spaced
    between ``high_freq`` and ``low_freq`` on an ERB scale.
    
    For a definition of ERB, see Moore, B. C. J., and Glasberg, B. R. (1983).
    "Suggested formulae for calculating auditory-filter bandwidths and
    excitation patterns," J. Acoust. Soc. Am. 74, 750-753.
    """
    return erb_point(
        low_freq,
        high_freq,
        np.arange(1, num + 1) / num
        )


def centre_freqs(fs, num_freqs, cutoff):
    """
    Calculates an array of centre frequencies (for :func:`make_erb_filters`)
    from a sampling frequency, lower cutoff frequency and the desired number of
    filters.
    
    :param fs: sampling rate
    :param num_freqs: number of centre frequencies to calculate
    :type num_freqs: int
    :param cutoff: lower cutoff frequency
    :return: same as :func:`erb_space`
    """
    return erb_space(cutoff, fs / 2, num_freqs)


def make_erb_filters(fs, centre_freqs, width=1.0):
    """
    This function computes the filter coefficients for a bank of 
    Gammatone filters. These filters were defined by Patterson and Holdworth for
    simulating the cochlea. 
    
    The result is returned as a :class:`ERBCoeffArray`. Each row of the
    filter arrays contains the coefficients for four second order filters. The
    transfer function for these four filters share the same denominator (poles)
    but have different numerators (zeros). All of these coefficients are
    assembled into one vector that the ERBFilterBank can take apart to implement
    the filter.
    
    The filter bank contains "numChannels" channels that extend from
    half the sampling rate (fs) to "lowFreq". Alternatively, if the numChannels
    input argument is a vector, then the values of this vector are taken to be
    the center frequency of each desired filter. (The lowFreq argument is
    ignored in this case.)
    
    Note this implementation fixes a problem in the original code by
    computing four separate second order filters. This avoids a big problem with
    round off errors in cases of very small cfs (100Hz) and large sample rates
    (44kHz). The problem is caused by roundoff error when a number of poles are
    combined, all very close to the unit circle. Small errors in the eigth order
    coefficient, are multiplied when the eigth root is taken to give the pole
    location. These small errors lead to poles outside the unit circle and
    instability. Thanks to Julius Smith for leading me to the proper
    explanation.
    
    Execute the following code to evaluate the frequency response of a 10
    channel filterbank::
    
        fcoefs = MakeERBFilters(16000,10,100);
        y = ERBFilterBank([1 zeros(1,511)], fcoefs);
        resp = 20*log10(abs(fft(y')));
        freqScale = (0:511)/512*16000;
        semilogx(freqScale(1:255),resp(1:255,:));
        axis([100 16000 -60 0])
        xlabel('Frequency (Hz)'); ylabel('Filter Response (dB)');
    
    | Rewritten by Malcolm Slaney@Interval.  June 11, 1998.
    | (c) 1998 Interval Research Corporation
    |
    | (c) 2012 Jason Heeris (Python implementation)
    """
    T = 1 / fs
    # Change the followFreqing three parameters if you wish to use a different
    # ERB scale. Must change in ERBSpace too.
    # TODO: factor these out
    ear_q = 9.26449 # Glasberg and Moore Parameters
    min_bw = 24.7
    order = 1

    erb = width*((centre_freqs / ear_q) ** order + min_bw ** order) ** ( 1 /order)
    B = 1.019 * 2 * np.pi * erb

    arg = 2 * centre_freqs * np.pi * T
    vec = np.exp(2j * arg)

    A0 = T
    A2 = 0
    B0 = 1
    B1 = -2 * np.cos(arg) / np.exp(B * T)
    B2 = np.exp(-2 * B * T)
    
    rt_pos = np.sqrt(3 + 2 ** 1.5)
    rt_neg = np.sqrt(3 - 2 ** 1.5)
    
    common = -T * np.exp(-(B * T))
    
    # TODO: This could be simplified to a matrix calculation involving the
    # constant first term and the alternating rt_pos/rt_neg and +/-1 second
    # terms
    k11 = np.cos(arg) + rt_pos * np.sin(arg)
    k12 = np.cos(arg) - rt_pos * np.sin(arg)
    k13 = np.cos(arg) + rt_neg * np.sin(arg)
    k14 = np.cos(arg) - rt_neg * np.sin(arg)

    A11 = common * k11
    A12 = common * k12
    A13 = common * k13
    A14 = common * k14

    gain_arg = np.exp(1j * arg - B * T)

    gain = np.abs(
            (vec - gain_arg * k11)
          * (vec - gain_arg * k12)
          * (vec - gain_arg * k13)
          * (vec - gain_arg * k14)
          * (  T * np.exp(B * T)
             / (-1 / np.exp(B * T) + 1 + vec * (1 - np.exp(B * T)))
            )**4
        )

    allfilts = np.ones_like(centre_freqs)
    
    fcoefs = np.column_stack([
        A0 * allfilts, A11, A12, A13, A14, A2*allfilts,
        B0 * allfilts, B1, B2,
        gain
    ])
    
    return fcoefs


def erb_filterbank(wave, coefs):
    """
    :param wave: input data (one dimensional sequence)
    :param coefs: gammatone filter coefficients
    
    Process an input waveform with a gammatone filter bank. This function takes
    a single sound vector, and returns an array of filter outputs, one channel
    per row.
    
    The fcoefs parameter, which completely specifies the Gammatone filterbank,
    should be designed with the :func:`make_erb_filters` function.
    
    | Malcolm Slaney @ Interval, June 11, 1998.
    | (c) 1998 Interval Research Corporation
    | Thanks to Alain de Cheveigne' for his suggestions and improvements.
    |
    | (c) 2013 Jason Heeris (Python implementation)
    """
    output = np.zeros((coefs[:,9].shape[0], wave.shape[0]))
    
    gain = coefs[:, 9]
    # A0, A11, A2
    As1 = coefs[:, (0, 1, 5)]
    # A0, A12, A2
    As2 = coefs[:, (0, 2, 5)]
    # A0, A13, A2
    As3 = coefs[:, (0, 3, 5)]
    # A0, A14, A2
    As4 = coefs[:, (0, 4, 5)]
    # B0, B1, B2
    Bs = coefs[:, 6:9]
    
    # Loop over channels
    for idx in range(0, coefs.shape[0]):
        # These seem to be reversed (in the sense of A/B order), but that's what
        # the original code did...
        # Replacing these with polynomial multiplications reduces both accuracy
        # and speed.
        y1 = sgn.lfilter(As1[idx], Bs[idx], wave)
        y2 = sgn.lfilter(As2[idx], Bs[idx], y1)
        y3 = sgn.lfilter(As3[idx], Bs[idx], y2)
        y4 = sgn.lfilter(As4[idx], Bs[idx], y3)
        output[idx, :] = y4 / gain[idx]
        
    return output


# -*- coding: utf-8 -*-
"""
Created on 2022/11/11
@project: SPAIC
@filename: AnimateScatter
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
import matplotlib.pyplot as plt
import matplotlib.animation as animation
import numpy as np

class AnimatedScatter(object):

    def __init__(self, positions, values):
        self.positions = positions
        self.values = values
        self.fig, self.ax = plt.subplots()
        self.ani = animation.FuncAnimation()

    def setup_plot(self):
        x, y, c = next(self.stream)
        self.scat = self.ax.scatter(x, y, c=c, animation=True)
        return self.scat

    def data_stream(self):
        pass

    def update(self, i):
        posi, color = next(self.stream)
        self.scat.set_offsets(posi)
        self.scat.set_array(color)
        return self.scat


    def show(self):
        plt.show()


# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: Conv_RSTDP.py
@time:2021/12/30 13:49
@description:
"""
from .Learner import Learner
from spaic.Network.BaseModule import Op
from spaic.Network.Connections import conv_connect
from spaic.IO.utils import im2col
import numpy as np
import torch

class Conv2d_RSTDP(Learner):
    """
    Reward-modulated STDP. Adapted from `(Florian 2007)
    <https://florian.io/papers/2007_Florian_Modulated_STDP.pdf>`.
    Args:
        lr (int or float): learning rate
        trainable: It can be network or neurongroups
    Attributes:
        tau_plus (int or float): Time constant for pre -synaptic firing trace determines the range of interspike intervals over which synaptic occur.
        tau_minus (int or float): Time constant for post-synaptic firing trace.
        a_plus (float): Learning rate for post-synaptic.
        a_minus (flaot): Learning rate for pre-synaptic.
    """

    def __init__(self, trainable=None, **kwargs):
        super(Conv2d_RSTDP, self).__init__(trainable=trainable)
        self.prefered_backend = ['pytorch']
        self.name = 'Conv2d_RSTDP'
        self.learning_rate = kwargs.get('lr', 0.1)

        self._tau_constant_variables = dict()
        self._tau_constant_variables['tau_plus'] = kwargs.get('tau_plus', 20.0)
        self._tau_constant_variables['tau_minus'] = kwargs.get('tau_minus', 20.0)

        self._constant_variables = dict()
        self._constant_variables['A_plus'] = kwargs.get('A_plus', 1.0)
        self._constant_variables['A_minus'] = kwargs.get('A_minus', -1.0)

    def weight_update(self, weight, eligibility, reward):
        """
        Conv2d_RSTDP learning rule for ``conv_connect`` subclass of ``Connection`` class.
        Args:
            weight : weight between pre and post neurongroup
            eligibility: a decaying memory of the relationships between the recent pairs of pre and postsynaptic spike pairs
            reward: reward signal
        """
        # Compute weight update based on the eligibility value of the past timestep.
        with torch.no_grad():
            assert reward.ndim <= 1, 'The reward for conv2d_rstdp should be a global reward'
            weight.add_(self.learning_rate * reward * eligibility)
            return weight

    def build(self, backend):
        super(Conv2d_RSTDP, self).build(backend)
        self.dt = backend.dt

        for (key, tau_var) in self._tau_constant_variables.items():
            tau_var = np.exp(-self.dt / tau_var)
            shape = ()
            self.variable_to_backend(key, shape, value=tau_var)

        for (key, var) in self._constant_variables.items():
            if isinstance(var, np.ndarray):
                if var.size > 1:
                    var_shape = var.shape
                    shape = (1, *var_shape)  # (1, shape)
                else:
                    shape = ()
            elif isinstance(var, list) or isinstance(var, tuple):
                if len(var) > 1:
                    var_len = len(var)
                    shape = (1, var_len)  # (1, shape)
                else:
                    shape = ()
            else:
                shape = ()
            self.variable_to_backend(key, shape, value=var)

        permute_name = 'conv2d_rstdp_permute_dim'
        permute_dim_value = [0, 2, 1]
        self.variable_to_backend(permute_name, shape=None, value=permute_dim_value, is_constant=True)
        reward_name = 'Output_Reward[updated]'

        # Traverse all trainable connections
        for conn in self.trainable_connections.values():
            if not isinstance(conn, conv_connect):
                raise ValueError('Conv2d_RSTDP can only modify the connection defined by conv_connect, not the %s'% str(type(conn)))
            preg = conn.pre
            postg = conn.post
            pre_name = conn.get_input_name(preg, postg)
            post_name = conn.get_group_name(postg, 'O')
            weight_name = conn.get_link_name(preg, postg, 'weight')
            out_channels, in_channels, kh, kw = backend._variables[weight_name].size()  # (out_channels, in_channels, kh, kw)
            padding = conn.padding
            stride = conn.stride

            # p_plus tracks the influence of presynaptic spikes
            p_plus_name = pre_name + '_{p_plus}'
            # p_minus tracks the influence of postsynaptic spikes
            p_minus_name = post_name + '_{p_minus}'
            eligibility_name = weight_name + '_{eligibility}'
            kh_name = weight_name + '_{kh}'
            kw_name = weight_name + '_{kw}'
            padding_name = weight_name + '_{padding}'
            stride_name = weight_name + '_{stride}'
            view_name = weight_name + '_{conv2d_rstdp_view_dim}'

            p_plus_value_temp = np.zeros(backend._variables[pre_name].shape)
            p_plus_value = im2col(p_plus_value_temp, kh, kw, stride, padding)
            p_minus_value_temp = np.zeros(backend._variables[post_name].shape)  # (batch_size, out_channels, height, width)
            p_minus_value = p_minus_value_temp.reshape(p_minus_value_temp.shape[0], p_minus_value_temp.shape[1], -1)
            view_dim_value = [out_channels, in_channels, kh, kw]

            self.variable_to_backend(p_plus_name, p_plus_value.shape, value=p_plus_value)
            self.variable_to_backend(p_minus_name, p_minus_value.shape, value=p_minus_value)
            self.variable_to_backend(eligibility_name, backend._variables[weight_name].shape, value=0.0)
            self.variable_to_backend(kh_name, shape=(), value=kh, is_constant=True)
            self.variable_to_backend(kw_name, shape=(), value=kw, is_constant=True)
            self.variable_to_backend(padding_name, shape=(), value=padding, is_constant=True)
            self.variable_to_backend(stride_name, shape=(), value=stride, is_constant=True)
            self.variable_to_backend(view_name, shape=None, value=view_dim_value, is_constant=True)

            sum_dim_name = 'conv2d_rstdp_sum'
            sum_dim_value = 0
            self.variable_to_backend(sum_dim_name, shape=(), value=sum_dim_value, is_constant=True)

            # Equations
            # pre_temp = im2col_indices(pre, kh, kw, padding, stride)
            # p_plus *= np.exp(-dt / tau_plus)
            # p_plus += A_plus * pre_temp
            # post_temp = post.view(batch_size, out_channels, -1)
            # p_minus *= np.exp(-dt / tau_minus)
            # p_minus += A_minus * post_temp
            # eligibility = torch.bmm(post, p_plus.permute((0, 2, 1))) + torch.bmm(p_minus, pre.permute((0, 2, 1)))
            # eligibility = torch.sum(eligibility, dim=0)
            # eligibility = eligibility.view([out_channels, in_channels, kh, kw])

            # Update p_plus values
            self.op_to_backend('pre_name_temp', 'im2col_indices', [pre_name, kh_name, kw_name, padding_name, stride_name])  # (batch_size, channels*kh*kw, height*width)
            # self.op_to_backend('p_plus_temp', 'var_mult', 'A_plus', 'pre_name_temp'))
            # self.op_to_backend(p_plus_name, 'var_linear', 'tau_plus', p_plus_name, 'p_plus_temp'))
            self.op_to_backend('p_plus_temp', 'var_mult', ['tau_plus', p_plus_name])
            self.op_to_backend(p_plus_name, 'var_linear', 'A_plus', 'pre_name_temp', 'p_plus_temp')

            # Update p_minus values
            self.op_to_backend('post_name_temp', 'conv2d_flatten', post_name)  # (batch_size, out_channels, height * width)
            # self.op_to_backend('p_minus_temp', 'var_mult', 'A_minus', 'post_name_temp'))
            # self.op_to_backend(p_minus_name, 'var_linear', 'tau_minus', p_minus_name, 'p_minus_temp'))
            self.op_to_backend('p_minus_temp', 'var_mult', ['tau_minus', p_minus_name])
            self.op_to_backend(p_minus_name, 'var_linear', ['A_minus', 'post_name_temp', 'p_minus_temp'])

            # Calculate point eligibility value
            self.op_to_backend('p_plus_permute', 'permute', [p_plus_name + '[updated]', permute_name])
            self.op_to_backend('pre_post', 'bmm', ['post_name_temp', 'p_plus_permute'])

            self.op_to_backend('pre_permute', 'permute', ['pre_name_temp', permute_name])
            self.op_to_backend('post_pre', 'bmm', [p_minus_name + '[updated]', 'pre_permute'])
            self.op_to_backend('eligibility_temp', ['add', 'pre_post', 'post_pre'])
            self.op_to_backend('eligibility_sum', 'reduce_sum', ['eligibility_temp', sum_dim_name])
            self.op_to_backend(eligibility_name, 'view', ['eligibility_sum', view_name])

            self.op_to_backend(weight_name, self.weight_update, [weight_name, eligibility_name, reward_name])

Learner.register('conv2d_rstdp', Conv2d_RSTDP)





"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: utils.py
@time:2021/4/1 14:33
@description:
"""
import os
os.environ['KMP_DUPLICATE_LIB_OK'] = 'True'
from random import shuffle
from shutil import copy
import numpy as np
import scipy.signal as signal
import scipy.io.wavfile as wav
from scipy.ndimage import maximum_filter
# import soundfile as sf
# import cv2
import math

import struct
from spaic.utils import plot, gtgram
import matplotlib.pyplot as plt

import scipy.io.wavfile as wav



'''
==================audio preprocess method==================
'''
def wav_file_resample(file_path, dest_sample=16e3):
    """
    对WAV文件进行resample的操作
    Args:
        file_path: 需要进行resample操作的wav文件的路径
        dest_sample:目标采样率
    Returns:
        resampled: 降采样后的数据
        dest_sample: 目标采样率
    """
    sample_rate, sound_signal = wav.read(file_path)
    signal_num = int((sound_signal.shape[0]) / sample_rate * dest_sample)
    resampled = signal.resample(sound_signal, signal_num)
    return resampled, dest_sample

def wav_file_cut(file_path, signal_num=16e3):
    """
    对WAV文件进行裁剪操作
    Args:
        file_path: 需要进行resample操作的wav文件的路径
        signal_num:目标数据数量
    Returns:
        cropped_data: 裁剪后的数据

    """
    sample_rate, sound_signal = wav.read(file_path)
    cropped_data = signal.resample(sound_signal, int(signal_num))
    return cropped_data

def _dataset_exists(root, class_labels):
    if os.path.exists(root):
        for cls_id in class_labels.keys():
            if not os.path.isdir(os.path.join(root, 'test', cls_id)):
                return False
            if not os.path.isdir(os.path.join(root, 'train', cls_id)):
                return False
        return True
    else:
        return False

def save_mfcc_feature(root, npz_name, sample_rate=16e3, signal_num=16e3, class_labels=None, **kwargs):
    from python_speech_features import mfcc
    if npz_name == '':
        feature_name = 'mfcc_feature'

    # set class labels
    if class_labels is None:
        raise ValueError('Missing class labels dict')
    else:
        classes = class_labels

    data = {
        'train_audios': [],
        'train_labels': [],
        'test_audios': [],
        'test_labels': [],
        'Time': [],
        'neuron_num': []
    }

    dataset_name = ['train', 'test']  # if is_train else 'test'
    for subset in dataset_name:
        for cls in classes.keys():
            cur_dir = os.path.join(root, subset, cls)
            for file in os.listdir(cur_dir):
                if not file.endswith('wav'):
                    continue
                wavform = wav_file_cut(os.path.join(cur_dir, file), signal_num)
                from python_speech_features import mfcc
                feature_mfcc = mfcc(wavform, samplerate=sample_rate)
                feature_mfcc = feature_mfcc.flatten()
                audios_name = "{}_audios".format(subset.lower())
                labels_name = "{}_labels".format(subset.lower())
                data[audios_name].append(feature_mfcc)
                data[labels_name].append(classes[cls])

    # 将音频数据存储为.npz文件
    data_root = os.path.join(root, feature_name)
    npz_name = feature_name + '.npz'
    data['neuron_num'] = len(data['train_audios'][0])
    trainMaxTime = get_Max(data['train_audios'])
    testMaxTime = get_Max(data['test_audios'])
    data['Time'] = max(trainMaxTime, testMaxTime)

    for k in data.keys():
        data[k] = np.array(data[k], dtype=object)
    np.savez(data_root, train_audios=data['train_audios'], train_labels=data['train_labels'],
             test_audios=data['test_audios'], test_labels=data['test_labels'], Time=data['Time'], neuron_num=data['neuron_num'])
    print(">> mfcc_features saved")
    return npz_name

def save_kp_feature(root=None, npz_name=None, sample_rate=16e3, class_labels=None, **kwargs):
    # parameters for extracting key points of audio
    window_size = kwargs.get('window_size', 0.016)
    stride = kwargs.get('stride', 0.008)
    kernels_num = kwargs.get('kernels_num', 100)
    freq_min = kwargs.get('freq_min', 20)
    Dr = kwargs.get('Dr', 3)
    Dc = kwargs.get('Dc', 3)
    significance_level = kwargs.get('significance_level', 3)
    if npz_name == '':
        feature_name = 'kp_feature'

    # set class labels
    if class_labels is None:
        raise ValueError('Missing class labels dict')
    else:
        classes = class_labels

    data = {
        'train_audios': [],
        'train_labels': [],
        'test_audios': [],
        'test_labels': [],
        'train_ids': [],
        'test_ids': [],
        'Time': [],
        'neuron_num': []
    }

    dataset_name = ['train', 'test']  # if is_train else 'test'
    for subset in dataset_name:
        for cls in classes.keys():
            cur_dir = os.path.join(root, subset, cls)
            for file in os.listdir(cur_dir):
                if not file.endswith('wav'):
                    continue

                wavform, sr = wav_file_resample(os.path.join(cur_dir, file), sample_rate)
                gmspec = fetchGmSpectrogram(wavform, sample_rate, window_size, stride, kernels_num, freq_min) # gtgram.gtgram(wavform, sample_rate, window_size, stride, kernels_num, freq_min, show=True)
                irow, icol, ival = extractKeyPoints(gmspec, Dr, Dc, significance_level)
                audios_name = "{}_audios".format(subset.lower())
                labels_name = "{}_labels".format(subset.lower())
                id_name = "{}_ids".format(subset.lower())
                data[audios_name].append(ival)
                data[labels_name].append(classes[cls])
                data[id_name].append(irow)

    # 将音频数据存储为.npz文件
    data_root = os.path.join(root, feature_name)
    npz_name = feature_name + '.npz'
    data['neuron_num'] = kernels_num
    trainMaxTime = get_Max(data['train_audios'])
    testMaxTime = get_Max(data['test_audios'])
    data['Time'] = max(trainMaxTime, testMaxTime)

    for k in data.keys():
        data[k] = np.array(data[k], dtype=object)
    np.savez(data_root, train_audios=data['train_audios'], train_labels=data['train_labels'], train_ids=data['train_ids'],
             test_audios=data['test_audios'], test_labels=data['test_labels'], test_ids=data['test_ids'], Time=data['Time'], neuron_num=data['neuron_num'])
    print(">> kp_feature saved")
    return npz_name

def load_kp_data(root, filename):
    data = {
        'train_audios': [],
        'test_audios': [],
        'train_ids': [],
        'test_ids': [],
        'train_labels': [],
        'test_labels': [],
        'Time': [],
        'neuron_num': []
    }
    fileroot = os.path.join(root, filename)
    data_temp = np.load(fileroot, allow_pickle=True)
    data['train_audios'] = data_temp['train_audios']
    data['test_audios'] = data_temp['test_audios']
    data['train_labels'] = data_temp['train_labels']
    data['test_labels'] = data_temp['test_labels']
    data['train_ids'] = data_temp['train_ids']
    data['test_ids'] = data_temp['test_ids']
    data['Time'] = data_temp['Time']
    data['neuron_num'] = data_temp['neuron_num']
    print(">> " + filename + " loaded")
    return data

def load_mfcc_data(root, filename):
    data = {
        'train_audios': [],
        'test_audios': [],
        'train_labels': [],
        'test_labels': [],
        'Time': [],
        'neuron_num': []
    }
    fileroot = os.path.join(root, filename)
    data_temp = np.load(fileroot, allow_pickle=True)
    data['train_audios'] = data_temp['train_audios']
    data['test_audios'] = data_temp['test_audios']
    data['train_labels'] = data_temp['train_labels']
    data['test_labels'] = data_temp['test_labels']
    data['Time'] = data_temp['Time']
    data['neuron_num'] = data_temp['neuron_num']
    print(">> " + filename + " loaded")
    return data

def dataset_split(source_root, target_root, ratio, is_shuffle):
    train_root = target_root + "\\" + 'train'
    test_root = target_root + "\\" + 'test'

    if not os.path.isdir(train_root):
        os.makedirs(train_root)
    if not os.path.isdir(test_root):
        os.makedirs(test_root)

    for class_name in os.listdir(source_root):
        class_root = os.path.join(source_root, class_name)

        train_dir = os.path.join(train_root, class_name)
        test_dir = os.path.join(test_root, class_name)
        if not os.path.isdir(train_dir):
            os.makedirs(train_dir)
        if not os.path.isdir(test_dir):
            os.makedirs(test_dir)

        samples = os.listdir(class_root)
        samples_len = len(samples)

        if is_shuffle is True:
            shuffle(samples)

        # i 用来计算文件数量
        i = 0
        to_path = train_dir
        for data_name in samples:
            split_num = math.ceil(ratio*samples_len)
            if i == 0:
                to_path = train_dir
            elif ((i % split_num) == 0):
                to_path = test_dir
            from_path = os.path.join(class_root, data_name)
            copy(from_path, to_path)
            i += 1

def reclassification(source_root, target_root, class_num, perperson_perclass_samplenum):
    '''
    将按录音者分类的digit语音数据集重保存为按录的音频的类别分类
    '''
    # 想保存到的根路径
    for i in range(class_num):
        save_dir = os.path.join(target_root, str(i))
        # 如果目录不存在，则创建
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)

    if os.path.exists(source_root):
        all_files = os.listdir(source_root)
    else:
        raise ValueError('The path'+ source_root +' is not exist')

    for file_name in all_files:
        file_root = os.path.join(source_root, file_name)
        samples = os.listdir(file_root)
        # i 用来计算文件数量，k 用来计算应存放到哪一类中
        i = 0
        k = 0
        for data_name in samples:
            if i == 0:
                k = 0
            elif ((i%perperson_perclass_samplenum)==0):
                k += 1
            from_path = os.path.join(file_root, data_name)
            to_path = target_root + "\\" + str(k)
            copy(from_path, to_path)
            i += 1

def datasetAlignment(source, maxNum):
    """
    Zeros are padded to each sample in the dataset according to the value of maxNum
    Args:
        source (ndarray): samples of dataset
        maxNum (int): The length of longest sample

    Returns:
        The data after padding

    """
    source = source.tolist()
    for n in range(len(source)):
        pad_len = maxNum-len(source[n])
        source[n] = np.pad(source[n], (0, pad_len), 'constant', constant_values=(0, 0))

    source = np.array(source)
    return source

def batchAlignment(source):
    source = source.tolist()
    maxNum = 0
    for n in range(len(source)):
        wav_sig = source[n]
        maxNum = max(maxNum, len(wav_sig))

    for n in range(len(source)):
        pad_len = maxNum-len(source[n])
        source[n] = np.pad(source[n], (0, pad_len), 'constant', constant_values=(0, 0))

    return source

def fetchGmSpectrogram(sig, fs=16e3, window_size=0.016, stride=0.008, kernels_num=32, freq_min=20, log=False, show=False):
    gmspec = gtgram.gtgram(sig, fs, window_size, stride, kernels_num, freq_min)
    if log:
        gmspec = np.log(gmspec)
    if show:
        p1 = plt.figure('spectrum', dpi=500)
        axes = p1.add_axes([0.1, 0.1, 0.9, 0.9])
        plot.gtgram_plot(gtgram.gtgram, axes, sig, fs, window_size, stride, kernels_num, freq_min)
        # plt.show()
        # print('')
    return gmspec

def extractKeyPoints(gmspec, Dr=13, Dc=13, significance_level=3):
    # print('begin extractKeyPoints')
    row_mask = np.ones([1, Dr])
    colum_mask = np.ones([Dc, 1])
    plus_mask = np.zeros([Dc, Dr])
    plus_mask[:, Dr >> 1] = 1
    plus_mask[Dc >> 1, :] = 1
    row_filter_spec = maximum_filter(gmspec, footprint=row_mask, mode='reflect')
    colum_filter_spec = maximum_filter(gmspec, footprint=colum_mask, mode='reflect')

    is_keypoint = np.logical_or((row_filter_spec == gmspec), (colum_filter_spec == gmspec))
    [irow, icol] = np.where(is_keypoint is True)
    ival = gmspec[irow, icol]
    # center_points = np.concatenate([irow,ival,])
    pad_gmspec = np.pad(gmspec, ((Dc >> 1, Dc >> 1), (Dr >> 1, Dr >> 1)), mode='symmetric')

    # todo: L默认填为矩形，其余位置由mask决定
    Lrow = np.empty([0, Dr])
    for i in range(irow.size):
        row_tmp = pad_gmspec[irow[i] + Dc >> 1, icol[i]:icol[i] + Dr].reshape([1, -1])
        Lrow = np.concatenate([Lrow, row_tmp], axis=0)
    Lcol = np.empty([Dc, 0])
    for i in range(icol.size):
        col_tmp = pad_gmspec[irow[i]:irow[i] + Dc, icol[i] + Dr >> 1].reshape([-1, 1])
        Lcol = np.concatenate([Lcol, col_tmp], axis=1)

    avg_Lrow = np.mean(Lrow, axis=1)
    avg_Lcol = np.mean(Lcol, axis=0)
    noise = np.zeros([ival.size])
    noise[avg_Lrow < avg_Lcol] = avg_Lrow[avg_Lrow < avg_Lcol]
    noise[avg_Lcol <= avg_Lrow] = avg_Lcol[avg_Lcol <= avg_Lrow]
    significant = ((ival - noise) > significance_level)

    # 去除噪点
    irow = irow[significant]
    icol = icol[significant]
    ival = ival[significant]
    print('..', sum(significant))
    return irow, icol, ival

def get_Max(data):
    '''
    get the maximum number of data
    Args:
        data (): can be spiking time or neuron ids

    Returns:

    '''
    maxData = 0
    for i in range(len(data)):
        tempMax = max(data[i])
        maxData = max(maxData, tempMax)
    return maxData
'''
==================image preprocess method==================
'''
def RGBtoGray(image):
    """
    Converts RGB image into gray image.

    Args:
        image: RGB image.
    Returns:
        Gray image.
    """
    import cv2
    return cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)

def GraytoBinary(image):
    """
    Converts input image into black and white (binary)

    Args:
        image: Gray image.
    Returns:
        Binary image.
    """
    import cv2
    return cv2.threshold(image, 0, 1, cv2.THRESH_BINARY)[1]


def reshape(image, shape):
    """
    Scale the image to (x, y).

    Args：
        image: Image to be rescaled.
        shape: Changed shape
    Returns:
        Re-scaled image.
    """
    import cv2
    return cv2.resize(image, shape)

def im2col(img, kh, kw, stride, padding='same'):
    '''
    :param img: 4D array
    :param kh: kernel_height
    :param kw: kernel_width
    :param stride:
    :param padding:
    :return:
    '''
    if padding == 'same':
        p1 = kh//2
        p2 = kw//2
        img = np.pad(img, ((0, 0), (0, 0), (p1, p1), (p2, p2),), 'constant')
    N, C, H, W = img.shape
    out_h = (H-kh)//stride[0]+1
    out_w = (W-kw)//stride[1]+1
    outsize = out_w*out_h
    col = np.empty((N, C, kw*kh, outsize,))
    for y in range(out_h):
        y_start = y * stride[0]
        y_end = y_start + kh
        for x in range(out_w):
            x_start = x * stride[1]
            x_end = x_start + kw
            col[:, :, 0:, y*out_w+x] = img[:, :, y_start:y_end, x_start:x_end].reshape(N, C, kh * kw)
    return col.reshape(N, -1, outsize)


def un_tar(file_name, output_root):
    # untar zip file to folder whose name is same as tar file
    import tarfile
    tar = tarfile.open(file_name)
    names = tar.getnames()

    file_name = os.path.basename(file_name)
    extract_dir = os.path.join(output_root, file_name.split('.')[0])

    # create folder if nessessary
    if os.path.isdir(extract_dir):
        pass
    else:
        os.makedirs(extract_dir)

    file_list = os.listdir(extract_dir)
    if len(file_list) == len(names):
        pass
    else:
        for name in names:
            tar.extract(name, extract_dir)
    tar.close()


'''
==================DVS preprocess method==================
'''
def load_aedat_v3(file_name: str):
    '''
    Args:
        file_name(str): path of the aedat v3 file
    Returns:
        a dict whose keys are ['t', 'x', 'y', 'p'] and values are ``numpy.ndarray``
    This function is written by referring to https://gitlab.com/inivation/dv/dv-python . It can be used for DVS128 Gesture.
    '''
    with open(file_name, 'rb') as bin_f:
        # skip ascii header
        line = bin_f.readline()
        while line.startswith(b'#'):
            if line == b'#!END-HEADER\r\n':
                break
            else:
                line = bin_f.readline()

        txyp = {
            't': [],
            'x': [],
            'y': [],
            'p': []
        }
        while True:
            header = bin_f.read(28)
            if not header or len(header) == 0:
                break

            # read header
            e_type = struct.unpack('H', header[0:2])[0]
            e_size = struct.unpack('I', header[4:8])[0]
            e_tsoverflow = struct.unpack('I', header[12:16])[0]
            e_capacity = struct.unpack('I', header[16:20])[0]

            data_length = e_capacity * e_size
            data = bin_f.read(data_length)
            counter = 0

            if e_type == 1:
                while data[counter:counter + e_size]:
                    aer_data = struct.unpack('I', data[counter:counter + 4])[0]
                    timestamp = struct.unpack('I', data[counter + 4:counter + 8])[0] | e_tsoverflow << 31
                    x = (aer_data >> 17) & 0x00007FFF
                    y = (aer_data >> 2) & 0x00007FFF
                    pol = (aer_data >> 1) & 0x00000001
                    counter = counter + e_size
                    txyp['x'].append(x)
                    txyp['y'].append(y)
                    txyp['t'].append(timestamp)
                    txyp['p'].append(pol)
            else:
                # non-polarity event packet, not implemented
                pass
        txyp['x'] = np.asarray(txyp['x'])
        txyp['y'] = np.asarray(txyp['y'])
        txyp['t'] = np.asarray(txyp['t'])
        txyp['p'] = np.asarray(txyp['p'])
        return txyp

def create_same_directory_structure(source_dir: str, target_dir: str) -> None:
    '''
    :param source_dir: Path of the directory that be copied from
    :type source_dir: str
    :param target_dir: Path of the directory that be copied to
    :type target_dir: str
    :return: None
    Create the same directory structure in ``target_dir`` with that of ``source_dir``.
    '''
    for sub_dir_name in os.listdir(source_dir):
        source_sub_dir = os.path.join(source_dir, sub_dir_name)
        if os.path.isdir(source_sub_dir):
            target_sub_dir = os.path.join(target_dir, sub_dir_name)
            os.mkdir(target_sub_dir)
            print(f'Mkdir [{target_sub_dir}].')
            create_same_directory_structure(source_sub_dir, target_sub_dir)


def integrate_events_file_to_frames_file_by_fixed_frames_number(events_np_file: str, output_dir: str, split_by: str, frames_num: int, H: int, W: int, print_save: bool = False) -> None:
    '''
    :param events_np_file: path of the events np file
    :type events_np_file: str
    :param output_dir: output directory for saving the frames
    :type output_dir: str
    :param split_by: 'time' or 'number'
    :type split_by: str
    :param frames_num: the number of frames
    :type frames_num: int
    :param H: the height of frame
    :type H: int
    :param W: the weight of frame
    :type W: int
    :param print_save: If ``True``, this function will print saved files' paths.
    :type print_save: bool
    :return: None
    Integrate a events file to frames by fixed frames number and save it. See ``cal_fixed_frames_number_segment_index`` and ``integrate_events_segment_to_frame`` for more details.
    '''
    fname = os.path.join(output_dir, os.path.basename(events_np_file))
    np.savez(fname, frames=integrate_events_by_fixed_frames_number(np.load(events_np_file), split_by, frames_num, H, W))
    if print_save:
        print(f'Frames [{fname}] saved.')

def integrate_events_by_fixed_frames_number(events: dict, split_by: str, frames_num: int, H: int, W: int) -> np.ndarray:
    '''
    :param events: a dict whose keys are ['t', 'x', 'y', 'p'] and values are ``numpy.ndarray``
    :type events: Dict
    :param split_by: 'time' or 'number'
    :type split_by: str
    :param frames_num: the number of frames
    :type frames_num: int
    :param H: the height of frame
    :type H: int
    :param W: the weight of frame
    :type W: int
    :return: frames
    :rtype: np.ndarray
    Integrate events to frames by fixed frames number. See ``cal_fixed_frames_number_segment_index`` and ``integrate_events_segment_to_frame`` for more details.
    '''
    j_l, j_r = cal_fixed_frames_number_segment_index(events['t'], split_by, frames_num)
    frames = np.zeros([frames_num, 2, H, W])
    for i in range(frames_num):
        frames[i] = integrate_events_segment_to_frame(events, H, W, j_l[i], j_r[i])
    return frames

def cal_fixed_frames_number_segment_index(events_t: np.ndarray, split_by: str, frames_num: int) -> tuple:
    '''
    :param events_t: events' t
    :type events_t: numpy.ndarray
    :param split_by: 'time' or 'number'
    :type split_by: str
    :param frames_num: the number of frames
    :type frames_num: int
    :return: a tuple ``(j_l, j_r)``
    :rtype: tuple
    Denote ``frames_num`` as :math:`M`, if ``split_by`` is ``'time'``, then
    .. math::
        \\Delta T & = [\\frac{t_{N-1} - t_{0}}{M}] \\\\
        j_{l} & = \\mathop{\\arg\\min}\\limits_{k} \\{t_{k} | t_{k} \\geq t_{0} + \\Delta T \\cdot j\\} \\\\
        j_{r} & = \\begin{cases} \\mathop{\\arg\\max}\\limits_{k} \\{t_{k} | t_{k} < t_{0} + \\Delta T \\cdot (j + 1)\\} + 1, & j <  M - 1 \\cr N, & j = M - 1 \\end{cases}
    If ``split_by`` is ``'number'``, then
    .. math::
        j_{l} & = [\\frac{N}{M}] \\cdot j \\\\
        j_{r} & = \\begin{cases} [\\frac{N}{M}] \\cdot (j + 1), & j <  M - 1 \\cr N, & j = M - 1 \\end{cases}
    '''
    j_l = np.zeros(shape=[frames_num], dtype=int)
    j_r = np.zeros(shape=[frames_num], dtype=int)
    N = events_t.size

    if split_by == 'number':
        di = N // frames_num
        for i in range(frames_num):
            j_l[i] = i * di
            j_r[i] = j_l[i] + di
        j_r[-1] = N

    elif split_by == 'time':
        dt = (events_t[-1] - events_t[0]) // frames_num
        idx = np.arange(N)
        for i in range(frames_num):
            t_l = dt * i + events_t[0]
            t_r = t_l + dt
            mask = np.logical_and(events_t >= t_l, events_t < t_r)
            idx_masked = idx[mask]
            j_l[i] = idx_masked[0]
            j_r[i] = idx_masked[-1] + 1

        j_r[-1] = N
    else:
        raise NotImplementedError

    return j_l, j_r

def integrate_events_segment_to_frame(events: dict, H: int, W: int, j_l: int = 0, j_r: int = -1) -> np.ndarray:
    '''
    :param events: a dict whose keys are ['t', 'x', 'y', 'p'] and values are ``numpy.ndarray``
    :type events: Dict
    :param H: height of the frame
    :type H: int
    :param W: weight of the frame
    :type W: int
    :param j_l: the start index of the integral interval, which is included
    :type j_l: int
    :param j_r: the right index of the integral interval, which is not included
    :type j_r:
    :return: frames
    :rtype: np.ndarray
    Denote a two channels frame as :math:`F` and a pixel at :math:`(p, x, y)` as :math:`F(p, x, y)`, the pixel value is integrated from the events data whose indices are in :math:`[j_{l}, j_{r})`:
.. math::
    F(p, x, y) &= \sum_{i = j_{l}}^{j_{r} - 1} \mathcal{I}_{p, x, y}(p_{i}, x_{i}, y_{i})
where :math:`\lfloor \cdot \rfloor` is the floor operation, :math:`\mathcal{I}_{p, x, y}(p_{i}, x_{i}, y_{i})` is an indicator function and it equals 1 only when :math:`(p, x, y) = (p_{i}, x_{i}, y_{i})`.
    '''
    # 累计脉冲需要用bitcount而不能直接相加，原因可参考下面的示例代码，以及
    # https://stackoverflow.com/questions/15973827/handling-of-duplicate-indices-in-numpy-assignments
    # We must use ``bincount`` rather than simply ``+``. See the following reference:
    # https://stackoverflow.com/questions/15973827/handling-of-duplicate-indices-in-numpy-assignments

    # Here is an example:

    # height = 3
    # width = 3
    # frames = np.zeros(shape=[2, height, width])
    # events = {
    #     'x': np.asarray([1, 2, 1, 1]),
    #     'y': np.asarray([1, 1, 1, 2]),
    #     'p': np.asarray([0, 1, 0, 1])
    # }
    #
    # frames[0, events['y'], events['x']] += (1 - events['p'])
    # frames[1, events['y'], events['x']] += events['p']
    # print('wrong accumulation\n', frames)
    #
    # frames = np.zeros(shape=[2, height, width])
    # for i in range(events['p'].__len__()):
    #     frames[events['p'][i], events['y'][i], events['x'][i]] += 1
    # print('correct accumulation\n', frames)
    #
    # frames = np.zeros(shape=[2, height, width])
    # frames = frames.reshape(2, -1)
    #
    # mask = [events['p'] == 0]
    # mask.append(np.logical_not(mask[0]))
    # for i in range(2):
    #     position = events['y'][mask[i]] * width + events['x'][mask[i]]
    #     events_number_per_pos = np.bincount(position)
    #     idx = np.arange(events_number_per_pos.size)
    #     frames[i][idx] += events_number_per_pos
    # frames = frames.reshape(2, height, width)
    # print('correct accumulation by bincount\n', frames)

    frame = np.zeros(shape=[2, H * W])
    x = events['x'][j_l: j_r].astype(int)  # avoid overflow
    y = events['y'][j_l: j_r].astype(int)
    p = events['p'][j_l: j_r]
    mask = []
    mask.append(p == 0)
    mask.append(np.logical_not(mask[0]))
    for c in range(2):
        position = y[mask[c]] * W + x[mask[c]]
        events_number_per_pos = np.bincount(position)
        frame[c][np.arange(events_number_per_pos.size)] += events_number_per_pos
    return frame.reshape((2, H, W))

def integrate_events_file_to_frames_file_by_fixed_duration(events_np_file: str, output_dir: str, duration: int, H: int, W: int, print_save: bool = False) -> None:
    '''
    :param events_np_file: path of the events np file
    :type events_np_file: str
    :param output_dir: output directory for saving the frames
    :type output_dir: str
    :param duration: the time duration of each frame
    :type duration: int
    :param H: the height of frame
    :type H: int
    :param W: the weight of frame
    :type W: int
    :param print_save: If ``True``, this function will print saved files' paths.
    :type print_save: bool
    :return: None
    Integrate events to frames by fixed time duration of each frame.
    '''
    frames = integrate_events_by_fixed_duration(np.load(events_np_file), duration, H, W)
    fname, _ = os.path.splitext(os.path.basename(events_np_file))
    fname = os.path.join(output_dir, f'{fname}_{frames.shape[0]}.npz')
    np.savez(fname, frames=frames)
    if print_save:
        print(f'Frames [{fname}] saved.')
    return frames.shape[0]

def integrate_events_by_fixed_duration(events: dict, duration: int, H: int, W: int) -> np.ndarray:
    '''
    :param events: a dict whose keys are ['t', 'x', 'y', 'p'] and values are ``numpy.ndarray``
    :type events: Dict
    :param duration: the time duration of each frame
    :type duration: int
    :param H: the height of frame
    :type H: int
    :param W: the weight of frame
    :type W: int
    :return: frames
    :rtype: np.ndarray
    Integrate events to frames by fixed time duration of each frame.
    '''
    t = events['t']
    N = t.size

    frames = []
    left = 0
    right = 0
    while True:
        t_l = t[left]
        while True:
            if right == N or t[right] - t_l > duration:
                break
            else:
                right += 1
        # integrate from index [left, right)
        frames.append(np.expand_dims(integrate_events_segment_to_frame(events, H, W, left, right), 0))

        left = right

        if right == N:
            return np.concatenate(frames)



# if __name__ == "__main__":
#     import numpy as np
#     import matplotlib.pyplot as plt
#     import os
#     import wave
#
#     # 读入音频。
#     path = r"F:\GitCode\Python\datasets\TidigitsWAV1\train\zero"
#     name = '1.wav'
#     # 我音频的路径为E:\SpeechWarehouse\zmkm\zmkm0.wav
#     filename = os.path.join(path, name)
#
#     # 打开语音文件。
#     f = wave.open(filename, 'rb')
#     # 得到语音参数
#     params = f.getparams()
#     nchannels, sampwidth, framerate, nframes = params[:4]
#     # ---------------------------------------------------------------#
#     # 将字符串格式的数据转成int型
#     print("reading wav file......")
#     strData = f.readframes(nframes)
#     waveData = np.fromstring(strData, dtype=np.short)
#     # 归一化
#     waveData = waveData * 1.0 / max(abs(waveData))
#     # 将音频信号规整乘每行一路通道信号的格式，即该矩阵一行为一个通道的采样点，共nchannels行
#     waveData = np.reshape(waveData, [nframes, nchannels]).T  # .T 表示转置
#     f.close()  # 关闭文件
#     print("file is closed!")
#     # ----------------------------------------------------------------#
#     '''绘制语音波形'''
#     print("plotting signal wave...")
#     time = np.arange(0, nframes) * (1.0 / framerate)  # 计算时间
#     time = np.reshape(time, [nframes, 1]).T
#     plt.plot(time[0, :nframes], waveData[0, :nframes], c="b")
#     plt.axis('off')  # no axis
#     plt.axes([0., 0., 1., 1.], frameon=False, xticks=[], yticks=[])
#     # plt.xlabel("time")
#     # plt.ylabel("amplitude")
#     # plt.title("Original wave")
#     plt.show()
#     print('end')


#     sroot = r'F:\GitCode\Python\datasets\TidigitsWAV'
#     troot = r'F:\GitCode\Python\datasets\TidigitsWAV1'
    # classes = {
    #     "zero": 0,
    #     "one": 1,
    #     "two": 2,
    #     "three": 3,
    #     "four": 4,
    #     "five": 5,
    #     "six": 6,
    #     "seven": 7,
    #     "eight": 8,
    #     "nine": 9,
    #     "oh": 10
    # }
    # save_mfcc_feature(root=troot, npz_name='mfcc_test.npz', class_labels=classes)
#     dataset_split(sroot, troot, 0.7, True)
#     print('end')




    # sroot = r'C:\Users\hp\Desktop\AudioMNIST'
    # troot = r'C:\Users\hp\Desktop\SpeechMNIST'
    # reclassification(sroot, troot, 10, 50)

    # time_now = time.time()
    # path = r'F:\GitCode\Python\dataset\AudioMNIST\train\0\0_01_2.wav'
    # sound, sampling_freq = wav_file_resample(path, 16e3)
    # print(time.time() - time_now)
    # time_now1 = time.time()
    # sound1, fs = librosa.load(path, sr=16e3)
    # print(time.time() - time_now1)

    # root = r'F:\GitCode\Python\datasets\AudioMNIST'
    # save_numpy_format(root, True, sample_rate=16e3)
    # save_numpy_format(root, False, sample_rate=16e3)

    # load_audio_data(root, True)
    # load_audio_data(root, False)
    # source = data['train_audios']
    # maxNum = data['maxNum']
    # datasetAlignment(source, maxNum)

    # root = r'F:\GitCode\Python\datasets\AudioMNIST'
    # filenameTr = 'train.npz'
    # root = r'F:\GitCode\Python\datasets\DigitsVoices'
    # filenameTr = 'train_kernels_num50significance_level10.npz'
    # filerootTr = os.path.join(root, filenameTr)
    # dataTr = np.load(filerootTr, allow_pickle=True)
    # trian_audios0 = dataTr['train_ids'][0]
    # trian_audios1 = dataTr['train_ids'][1]
    # trian_audios2 = dataTr['train_ids'][2]
    # trian_audios3 = dataTr['train_ids'][3]
    # train_ids = dataTr['train_ids']
    # maxvalue = get_Max(dataTr['train_audios'])
    # filenameTe = 'test_kernels_num50significance_level10.npz'
    # filerootTe = os.path.join(root, filenameTe)
    # dataTe = np.load(filerootTe, allow_pickle=True)
    # test_audios0 = dataTe['test_audios'][0]
    # test_audios1 = dataTe['test_audios'][1]
    # test_audios2 = dataTe['test_audios'][2]
    # test_audios3 = dataTe['test_audios'][3]
    # test_audios = dataTe['test_audios']
    # test_ids = dataTe['test_ids']
    # maxvalue = get_Max(dataTe['test_audios'])
    # print(maxvalue)
    # print('')




# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: Synapse.py
@time:2022/5/26 9:19
@description:
"""
from ..Network.Topology import SynapseModel
from ..Neuron.Neuron import NeuronGroup
import numpy as np

class Basic_synapse(SynapseModel):
    """
    Basic synapse
    Compute Isyn
    """

    def __init__(self, conn, **kwargs):
        super(Basic_synapse, self).__init__(conn)



        if conn.is_sparse:
            self._syn_operations.append([conn.post_var_name + '[post]', 'sparse_mat_mult_weight', 'weight[link]',
                                         self.input_name])
        elif  'complex' in conn.post.model_name and conn.post.model_name != 'double_complex':
            if conn.max_delay > 0:
                self._syn_operations.append(
                    [conn.post_var_name + '[post]', 'mat_mult_weight_complex', self.input_name,
                     'weight[link]', 'complex_beta[post]', 'delay[link]'])
            else:
                self._syn_operations.append(
                    [conn.post_var_name + '[post]', 'mat_mult_weight_complex', self.input_name,
                     'weight[link]', 'complex_beta[post]'])
        elif conn.post.model_name == 'double_complex':
            if conn.max_delay > 0:
                self._syn_operations.append(
                    [conn.post_var_name + '[post]', 'mat_mult_weight_2complex', self.input_name,
                     'weight[link]', 'complex_beta[post]', 'delay[link]'])
            else:
                self._syn_operations.append(
                    [conn.post_var_name + '[post]', 'mat_mult_weight_2complex', self.input_name,
                     'weight[link]', 'complex_beta[post]'])
        elif conn.max_delay > 0:
            self._syn_operations.append(
                [conn.post_var_name + '[post]', 'mult_sum_weight', self.input_name, 'weight[link]'])
        else:
            self._syn_operations.append(
                [conn.post_var_name + '[post]', 'mat_mult_weight', self.input_name,
                 'weight[link]'])

        if conn.bias_flag:
            self._syn_operations.append([conn.post_var_name + '[post]', 'add', conn.post_var_name + '[post]', 'bias[link]'])

# SynapseModel.register('basic_synapse', Basic_synapse)
SynapseModel.register('basic', Basic_synapse)

class conv_synapse(SynapseModel):
    """
    conv synapse
    Compute Isyn
    """

    def __init__(self, conn, **kwargs):
        super(conv_synapse, self).__init__(conn)
        if conn.post.model_name == 'complex':
            self._syn_operations.append(
                [conn.post_var_name + '[post]', 'conv_2d_complex', self.input_name, 'weight[link]',
                 'stride[link]', 'padding[link]', 'dilation[link]', 'groups[link]', 'complex_beta[post]'])
        else:
            self._syn_operations.append(
                [conn.post_var_name + '[post]', 'conv_2d', self.input_name, 'weight[link]',
                 'stride[link]', 'padding[link]', 'dilation[link]', 'groups[link]']) # every time a new parameter will be added to all conditions, that's silly

        if conn.bias_flag:
            self._syn_operations.append(
                [conn.post_var_name + '[post]', 'conv_add_bias', conn.post_var_name + '[post]', 'bias[link]'])

# SynapseModel.register('conv_synapse', conv_synapse)
SynapseModel.register('conv', conv_synapse)


class ConvTranspose_synapse(SynapseModel):

    def __init__(self, conn, **kwargs):
        super(ConvTranspose_synapse, self).__init__(conn)

        self._syn_operations.append(
            [conn.post_var_name + '[post]', 'conv_trans2d', self.input_name, 'weight[link]',
             'stride[link]', 'padding[link]', 'dilation[link]', 'groups[link]'])

        if conn.bias_flag:
            raise ValueError("bias for conv_transpose is not supported")

    # SynapseModel.register('conv_synapse', conv_synapse)


SynapseModel.register('conv_transpose', ConvTranspose_synapse)

class DirectPass_synapse(SynapseModel):
    """
    DirectPass synapse
    target_name = input_name
    """

    def __init__(self, conn, **kwargs):
        super(DirectPass_synapse, self).__init__(conn)
        self._syn_operations.append([conn.post_var_name + '[post]', 'assign', self.input_name])

# SynapseModel.register('directpass_synapse', DirectPass_synapse)
SynapseModel.register('directpass', DirectPass_synapse)

class Dropout_synapse(SynapseModel):
    """
    During training, randomly zeroes some of the elements of the input tensor with probability :
    attr:`p` using samples from a Bernoulli distribution. Each channel will be zeroed out independently on every forward call.
    Args:
        p: probability of an element to be zeroed. Default: 0.5
        inplace: If set to ``True``, will do this operation in-place. Default: ``False``
    """

    def __init__(self, conn, **kwargs):
        super(Dropout_synapse, self).__init__(conn)
        self._syn_constant_variables['p'] = conn.parameters.get('p', 0.5)
        self._syn_constant_variables['inplace'] = conn.parameters.get('inplace', False)
        self._syn_operations.append([conn.pre_var_name + '[input]', 'dropout', self.input_name, 'p',
                                     'inplace'])

# SynapseModel.register('dropout_synapse', Dropout_synapse)
SynapseModel.register('dropout', Dropout_synapse)

class AvgPool_synapse(SynapseModel):

    def __init__(self, conn, **kwargs):
        super(AvgPool_synapse, self).__init__(conn)
        if conn.pool_only:
            self._syn_operations.append([conn.post_var_name + '[post]', 'avg_pool2d', self.input_name,
                                         'pool_kernel_size[link]', 'pool_stride[link]', 'pool_padding[link]'])
        else:
            if conn.pool_before:
                # when pooling before, the return operator name should be pre_var_name
                self._syn_operations.append([conn.pre_var_name + '[input]', 'avg_pool2d', self.input_name,
                                             'pool_kernel_size[link]', 'pool_stride[link]', 'pool_padding[link]'])
            else:
                self._syn_operations.append([conn.post_var_name + '[post]', 'avg_pool2d', conn.post_var_name + '[post]',
                                             'pool_kernel_size[link]', 'pool_stride[link]', 'pool_padding[link]'])

# SynapseModel.register('avgpool_synapse', AvgPool_synapse)
SynapseModel.register('avgpool', AvgPool_synapse)

class MaxPool_synapse(SynapseModel):

    def __init__(self, conn, **kwargs):
        super(MaxPool_synapse, self).__init__(conn)
        if conn.pool_only:
            self._syn_operations.append([conn.post_var_name + '[post]', 'avg_pool2d', self.input_name,
                                         'pool_kernel_size[link]', 'pool_stride[link]', 'pool_padding[link]'])
        else:
            if conn.pool_before:
                # when pooling before, the return operator name should be pre_var_name
                self._syn_operations.append([conn.pre_var_name + '[input]', 'max_pool2d', self.input_name,
                                             'pool_kernel_size[link]', 'pool_stride[link]', 'pool_padding[link]'])
            else:
                if conn.post.model_name == 'complex':
                    self._syn_operations.append(
                        [conn.post_var_name + '[post]', 'post_max_pool2d_complex', conn.post_var_name + '[post]',
                         'pool_kernel_size[link]', 'pool_stride[link]', 'pool_padding[link]'])
                else:
                    self._syn_operations.append(
                        [conn.post_var_name + '[post]', 'max_pool2d', conn.post_var_name + '[post]',
                         'pool_kernel_size[link]', 'pool_stride[link]', 'pool_padding[link]'])

# SynapseModel.register('maxpool_synapse', MaxPool_synapse)
SynapseModel.register('maxpool', MaxPool_synapse)

class Upsample_synapse(SynapseModel):
    def __init__(self, conn, **kwargs):
        super(Upsample_synapse, self).__init__(conn)
        self._syn_operations.append([conn.pre_var_name + '[input]', 'upsample', self.input_name, 'upscale[link]'])

SynapseModel.register('upsample', Upsample_synapse)

class BatchNorm2d_synapse(SynapseModel):

    def __init__(self, conn, **kwargs):
        super(BatchNorm2d_synapse, self).__init__(conn)
        if 'num_features' in conn.syn_kwargs.keys():
            num_features = conn.syn_kwargs['num_features']
        else:
            raise ValueError('The parameter num_features is not given.')

        if 'num_features' in conn.syn_kwargs.keys():
            num_features = conn.syn_kwargs['num_features']
        else:
            raise ValueError('The parameter num_features is not given. Set num_features in the syn_kwargs dict '
                             'initialization parameter of the Connecion class.')

        self._syn_constant_variables['num_features'] = num_features
        self._syn_operations.append([conn.post_var_name + '[post]', 'batchnorm2d', conn.post_var_name + '[post]',
                                     'num_features'])

# SynapseModel.register('BatchNorm2d_synapse', BatchNorm2d_synapse)
SynapseModel.register('batchnorm2d', BatchNorm2d_synapse)

class Flatten_synapse(SynapseModel):

    def __init__(self, conn, **kwargs):
        super(Flatten_synapse, self).__init__(conn)
        self._syn_constant_variables['view_dim'] = [-1, conn.pre_num]
        self._syn_operations.append([conn.pre_var_name + '[input]', 'view', self.input_name,
                                     'view_dim'])

# SynapseModel.register('flatten_synapse', Flatten_synapse)
SynapseModel.register('flatten', Flatten_synapse)


class Electrical_synapse(SynapseModel):
    """
    Electrical synapse
    Iele = weight *（V(l-1) - V(l)）
    """

    def __init__(self, conn, **kwargs):
        super(Electrical_synapse, self).__init__(conn)
        # V_post = conn.get_post_name(conn.post, 'V')
        # V_pre = conn.get_pre_name(conn.pre, 'V')
        # Vtemp_post = conn.get_link_name(conn.pre, conn.post, 'Vtemp')
        # I_post = conn.get_post_name(conn.post, 'I_ele')
        # weight = conn.get_link_name(conn.pre, conn.post, 'weight')
        # Vtemp_pre = conn.get_link_name(conn.post, conn.pre, 'Vtemp')
        # I_pre = conn.get_pre_name(conn.pre, 'I_ele')
        #
        # self._syn_variables[Vtemp_post] = 0.0
        # self._syn_variables[I_post] = 0.0
        # self._syn_variables[Vtemp_pre] = 0.0
        # self._syn_variables[I_pre] = 0.0
        # self._syn_operations.append([Vtemp_post, 'minus', V_pre, V_post])
        # self._syn_operations.append([I_post, 'var_mult', weight, Vtemp_post + '[updated]'])
        # self._syn_operations.append([Vtemp_pre, 'minus', V_post, V_pre])
        # self._syn_operations.append([I_pre, 'var_mult', weight, Vtemp_pre + '[updated]'])

        # self._syn_variables['Vprepost'] = np.zeros([1, conn.pre_num, conn.post_num])
        assert isinstance(conn.pre, NeuronGroup) and isinstance(conn.post,
                                                                NeuronGroup), f"Electrical synapses exist in connections in which the presynaptic and postsynaptic objects are neurongroups"

        self._syn_variables['Isyn[post]'] = np.zeros([1, conn.post_num])
        self._syn_variables['Isyn[pre]'] = np.zeros([1, conn.pre_num])
        self._syn_constant_variables['unsequence_dim'] = 0
        self._syn_constant_variables['permute_dim'] = [1, 2, 0]
        self._syn_constant_variables['Vpre_permute_dim'] = [2, 1, 0]
        self._syn_constant_variables['post_sum_dim'] = 2
        self._syn_constant_variables['pre_sum_dim'] = 1

        # unsequence_dim_name =
        self._syn_operations.append(['Vpre', 'unsqueeze', 'V[pre]', 'unsequence_dim'])
        self._syn_operations.append(['Vpre_temp', 'permute', 'Vpre', 'Vpre_permute_dim'])
        self._syn_operations.append(['Vpost_temp', 'unsqueeze', 'V[post]', 'unsequence_dim'])
        # [pre_num, batch_size, post_num] [pre_num, batch_size, 1] [1, batch_size, post_num]
        self._syn_operations.append(['Vprepost', 'minus', 'Vpre_temp', 'Vpost_temp'])
        # [batch_size, post_num, pre_num]
        self._syn_operations.append(['Vprepost_temp', 'permute', 'Vprepost', 'permute_dim'])
        self._syn_operations.append(['I_post_temp', 'var_mult', 'Vprepost_temp', 'weight[link]'])
        # [batch_size, post_num]
        self._syn_operations.append(['Isyn[post]', 'reduce_sum', 'I_post_temp', 'post_sum_dim'])

        # [pre_num, batch_size, post_num]  [1, batch_size, post_num] [pre_num, batch_size, 1]
        self._syn_operations.append(['Vpostpre', 'minus', 'Vpost_temp', 'Vpre_temp'])
        # [batch_size, post_num, pre_num]
        self._syn_operations.append(['Vpostpre_temp', 'permute', 'Vpostpre', 'permute_dim'])
        self._syn_operations.append(['I_pre_temp', 'var_mult', 'Vpostpre_temp', 'weight[link]'])
        # [batch_size, pre_num]
        self._syn_operations.append(['Isyn[pre]', 'reduce_sum', 'I_pre_temp', 'pre_sum_dim'])


# SynapseModel.register('electrical_synapse', Electrical_synapse)
SynapseModel.register('electrical', Electrical_synapse)

class First_order_chemical_synapse(SynapseModel):
    """
    .. math:: Isyn(t) = weight * e^{-t/tau}
    """
    def __init__(self, conn, **kwargs):
        super(First_order_chemical_synapse, self).__init__(conn)
        from ..Network.Connections import FullConnection
        assert isinstance(conn, FullConnection)
        self._syn_tau_variables['tau[link]'] = kwargs.get('tau', 2.0)
        self._syn_variables['R[link]'] = np.zeros([1, conn.post_num])
        self._syn_variables['WgtSum[link]'] = np.zeros([1, conn.post_num])

        if conn.post.model_name == 'complex':
            self._syn_operations.append(
                ['WgtSum[link]', 'mat_mult_weight_complex', conn.pre_var_name + '[input][updated]',
                 'weight[link]', 'complex_beta[post]'])
        else:
            self._syn_operations.append(['WgtSum[link]', 'mat_mult_weight', '[input]', 'weight[link]'])

        self._syn_operations.append(['R[link]', 'var_linear', 'tau[link]', 'R[link]', 'WgtSum[link][updated]'])
        self._syn_operations.append([conn.post_var_name + '[post]', 'assign', 'R[link][updated]'])


SynapseModel.register('1_order_synapse', First_order_chemical_synapse)


class Second_order_chemical_synapse(SynapseModel):
    """
    .. math:: Isyn(t) = weight*( e^{-t/tau_r}} - e^{-t/tau_d} )
    """

    def __init__(self, conn, **kwargs):
        super(Second_order_chemical_synapse, self).__init__(conn)
        from ..Network.Connections import FullConnection
        assert isinstance(conn, FullConnection)
        self._syn_tau_variables['tau_r[link]'] = kwargs.get('tau_r', 9.0)
        self._syn_tau_variables['tau_d[link]'] = kwargs.get('tau_d', 2.0)
        self._syn_variables['R[link]'] = np.zeros([1, conn.post_num])
        self._syn_variables['D[link]'] = np.zeros([1, conn.post_num])
        self._syn_variables['WgtSum[link]'] = np.zeros([1, conn.post_num])
        if conn.post.model_name == 'complex':
            self._syn_operations.append(
                ['WgtSum[link]', 'mat_mult_weight_complex', conn.pre_var_name + '[input][updated]',
                 'weight[link]', 'complex_beta[post]'])
        else:
            self._syn_operations.append(['WgtSum[link]', 'mat_mult_weight', '[input]', 'weight[link]'])
        self._syn_operations.append(['R[link]', 'var_linear', 'tau_r[link]', 'R[link]', 'WgtSum[link][updated]'])
        self._syn_operations.append(['D[link]', 'var_linear', 'tau_d[link]', 'D[link]', 'WgtSum[link][updated]'])
        self._syn_operations.append([conn.post_var_name + '[post]', 'minus', 'R[link][updated]', 'D[link][updated]'])

SynapseModel.register('2_order_synapse', Second_order_chemical_synapse)

import torch
class Delayed_complex_synapse(SynapseModel):
    def __init__(self, conn, **kwargs):
        super(Delayed_complex_synapse, self).__init__(conn)
        self._backend = conn._backend
        self.delay = kwargs.get('delay', None)
        self.syn_tau = kwargs.get('tau', 10.0)
        self.syn_rotk = kwargs.get('rot_k', 6.0)
        self.delay_len = kwargs.get('delay_len', 1)
        self.max_delay_len = kwargs.get('max_delay_len', None)
        self.max_delay = kwargs.get('max_delay', 5.0)
        self.min_delay = kwargs.get('min_delay', 0.0)
        self.real_output = kwargs.get('real_output', True)
        self.weight_shape = list(conn.get_value('weight').shape)
        self.delay_shape = [self.delay_len] + self.weight_shape
        self.delay_buffer = torch.zeros(self.delay_shape, dtype=torch.cfloat).unsqueeze(0)
        self.Isyn = torch.zeros(self.weight_shape[0], dtype=torch.cfloat).unsqueeze(0)

        beta_s = np.exp(-self.dt/ self.syn_tau)
        rot_s = -self.dt*np.pi / (self.syn_rotk * self.syn_tau)
        # tmax = np.arctan(-self.syn_tau*rot_s)/-rot_s
        # vmax = np.exp(-tmax/self.syn_tau)*np.sin(-rot_s*tmax)
        self.v0 = 1.0/self.syn_tau
        self.complex_decay = torch.view_as_complex(torch.tensor([beta_s * np.cos(rot_s), beta_s * np.sin(rot_s)]))

        if self.delay is None:
            self.delay = self.min_delay + (self.max_delay-self.min_delay)*torch.rand(self.weight_shape, dtype=torch.float)
        elif self.max_delay is None:
            self.delay = torch.tensor(self.delay)
            self.max_delay = torch.amax(self.delay).item()
        if self.max_delay > self.syn_tau*self.syn_rotk:
            rot_d = -self.dt *np.pi / self.max_delay
        else:
            rot_d = rot_s
        self.complex_delay = torch.view_as_complex(torch.tensor([np.cos(rot_d), np.sin(rot_d)]))

        init_rot = -self.delay * rot_d / self.dt
        self.init_roting_complex = torch.view_as_complex(torch.stack([torch.cos(init_rot), torch.sin(init_rot)], dim=-1))

        self._syn_variables['delay_buffer[link]'] = self.delay_buffer
        self._syn_variables['Isyn[link]'] = self.Isyn
        self._syn_variables['init_roting_complex[link]'] = self.init_roting_complex
        self._syn_variables['complex_decay[link]'] = self.complex_decay
        self._syn_variables['complex_delay[link]'] = self.complex_delay

        if self.real_output:
            self._syn_operations.append([[conn.post_var_name+'[post]', 'delay_buffer[link]', 'Isyn[link]'], self.update_real_out,
                                         ['[input]','delay_buffer[link]', 'Isyn[link]', 'weight[link]', 'init_roting_complex[link]',
                                          'complex_delay[link]', 'complex_decay[link]']])
        else:
            self._syn_operations.append([[conn.post_var_name+'[post]', 'delay_buffer[link]', 'Isyn[link]'], self.update_complex_out,
                                         ['[input]','delay_buffer[link]', 'Isyn[link]', 'weight[link]', 'init_roting_complex[link]',
                                          'complex_delay[link]', 'complex_decay[link]']])


    @property
    def dt(self):
        return self._backend.dt


    def push_spike(self, spike, delay_buffer, init_roting_complex):
        if spike.dtype is torch.cfloat:
            vspk = spike.abs()
        else:
            vspk = spike

        vspk = torch.sum(vspk)
        if vspk > 0:
            if spike.dtype is not torch.cfloat:
                zero_real = torch.zeros_like(spike)
                spike = torch.view_as_complex(torch.stack((zero_real, spike), dim=-1))
            spike = spike.unsqueeze(-2)
            spike = self.v0*spike * init_roting_complex

            values, indices = torch.min(delay_buffer.abs(), dim=1)
            indices = indices.unsqueeze(1)
            spike = spike.unsqueeze(1)
            if torch.sum(values) > 0:
                delay_buffer = torch.cat([spike, delay_buffer], dim=1)
            else:
                delay_buffer.scatter_(dim=1, index=indices, src=spike)
        return delay_buffer

    def update_real_out(self, spike, delay_buffer, Isyn, weight, init_roting_complex, complex_delay, complex_decay):
        # push input spikes
        if spike.dtype is torch.cfloat:
            vspk = spike.abs()
        else:
            vspk = spike

        vspk = torch.sum(vspk)
        if vspk > 0:
            if spike.dtype is not torch.cfloat:
                zero_real = torch.zeros_like(spike)
                spike = torch.view_as_complex(torch.stack((zero_real, spike), dim=-1))
            spike = spike.unsqueeze(-2)
            spike = self.v0*spike * init_roting_complex

            values, indices = torch.min(delay_buffer.abs(), dim=1)
            indices = indices.unsqueeze(1)
            spike = spike.unsqueeze(1)
            if torch.sum(values) > 0:
                delay_buffer = torch.cat([spike, delay_buffer], dim=1)
            else:
                delay_buffer.scatter_(dim=1, index=indices, src=spike)

        # update and output
        delay_buffer = delay_buffer*complex_delay
        spike_mask = delay_buffer.real.gt(0.0)
        Isyn = Isyn*complex_decay + torch.sum(torch.sum(spike_mask*delay_buffer, 1)*weight, dim=-1)
        delay_buffer[spike_mask] = 0.0
        out = Isyn.real
        return out, delay_buffer, Isyn

    def update_complex_out(self, spike, delay_buffer, Isyn, weight, init_roting_complex, complex_delay, complex_decay):
        # push input spikes
        if spike.dtype is torch.cfloat:
            vspk = spike.abs()
        else:
            vspk = spike

        vspk = torch.sum(vspk)
        if vspk > 0:
            if spike.dtype is not torch.cfloat:
                zero_real = torch.zeros_like(spike)
                spike = torch.view_as_complex(torch.stack((zero_real, spike), dim=-1))
            spike = spike.unsqueeze(-2)
            spike = self.v0*spike * init_roting_complex

            values, indices = torch.min(delay_buffer.abs(), dim=1)
            indices = indices.unsqueeze(1)
            spike = spike.unsqueeze(1)
            if torch.sum(values) > 0:
                delay_buffer = torch.cat([spike, delay_buffer], dim=1)
            else:
                delay_buffer.scatter_(dim=1, index=indices, src=spike)

        # update and output
        delay_buffer = delay_buffer*complex_delay
        spike_mask = delay_buffer.real.gt(0.0)
        Isyn = Isyn*complex_decay + torch.sum(torch.sum(spike_mask*delay_buffer, 1)*weight, dim=-1)
        delay_buffer[spike_mask] = 0.0
        return Isyn, delay_buffer, Isyn

SynapseModel.register('delay_complex_synapse', Delayed_complex_synapse)

class Mix_order_chemical_synapse(SynapseModel):

    def __init__(self, conn=None, **kwargs):
        super(Mix_order_chemical_synapse, self).__init__(conn)
        if conn is not None:
            from ..Network.Connections import FullConnection
            assert isinstance(conn, FullConnection)
            assert conn.post.model_name == 'complex'
            tau_r = kwargs.get('tau_r', 2.0)
            tau_s = kwargs.get('tau_s', 25.0)
            assert tau_r > 0 and tau_s > 0
            self._syn_tau_variables['tau_r[link]'] = tau_r
            self._syn_tau_variables['tau_s[link]'] = tau_s
            self._syn_constant_variables['alpha_r[link]'] = 1.0 / tau_r
            self._syn_constant_variables['alpha_s[link]'] = 1.0 / tau_s
            self._syn_variables['R[link]'] = np.zeros([1, conn.post_num])
            self._syn_variables['S[link]'] = np.zeros([1, conn.post_num])
            self._syn_operations.append([[conn.post_var_name + '[post]', 'R[link]', 'S[link]'], self.update,
                                        [conn.pre_var_name + '[input][updated]', 'weight[link]', 'complex_beta[post]',
                                         'R[link]', 'S[link]', 'tau_r[link]','tau_s[link]','alpha_r[link]',
                                         'alpha_s[link]','[dt]']])


    def update(self, inp, weight, complex_beta, R, S, beta_r, beta_s, alpha_r, alpha_s, dt):
        if inp.dtype.is_complex:
            x = inp.unsqueeze(-2)
            complex_beta = complex_beta.unsqueeze(-1)
            rate = x.real
            time = x.imag
            O = complex_beta ** time * (rate * (0 + 1.0j))
            ratio = rate*rate
            WgtSumR = torch.sum(ratio * O * weight, dim=-1)
            WgtSumS = torch.sum((1-ratio) * O * weight, dim=-1)
        else:
            weight = weight.permute(1, 0)
            WgtSumR = torch.matmul(inp, weight)
            WgtSumR = WgtSumR*(0.0+1.0j)
            WgtSumS = 0.0

        S = beta_s * S + alpha_s*WgtSumS
        R = beta_r*R + alpha_r*WgtSumR + alpha_r*S*dt
        return R, R, S
SynapseModel.register('mix_order_synapse', Mix_order_chemical_synapse)








# -*- coding: utf-8 -*-
"""
Created on 2020/10/23
@project: SPAIC
@filename: SignalGenerator
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from spaic.IO.Pipeline import Pipline

class SigGenerator(Pipline):
    '''
    Generate signals such as sin/cos wave, random process...
    '''

    def __init__(self):
        super(SigGenerator, self).__init__()
        pass



# -*- coding: utf-8 -*-
"""
Created on 2020/12/23
@project: SPAIC
@filename: WithBuildFuncs
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description: 
"""
from spaic.Network.Assembly import Assembly
import spaic

# ================= Assembly Functions used in with ==========================

def add_assembly(name, assembly):
    context = spaic.global_assembly_context_list
    assert len(context) >0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    self.add_assembly(name, assembly)
    spaic.global_assembly_init_count -= 1



def del_assembly(assembly=None, name=None):

    context = spaic.global_assembly_context_list
    assert len(context) >0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    self.del_assembly(name, assembly)
    spaic.global_assembly_init_count -= 1


def add_connection(name, connection):
    context = spaic.global_assembly_context_list
    assert len(context) >0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    self.add_connection(name, connection)
    spaic.global_assembly_init_count -= 1


def del_connection(connection=None, name=None):

    context = spaic.global_assembly_context_list
    assert len(context) >0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    self.del_connection(connection, name)
    spaic.global_assembly_init_count -= 1



def copy_assembly(name, assembly):
    context = spaic.global_assembly_context_list
    assert len(context) > 0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    self.copy_assembly(name, assembly)
    spaic.global_assembly_init_count -= 1


def replace_assembly(old_assembly, new_assembly):
    context = spaic.global_assembly_context_list
    assert len(context) > 0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    self.replace_assembly(old_assembly, new_assembly)
    spaic.global_assembly_init_count -= 1


def merge_assembly(assembly):
    context = spaic.global_assembly_context_list
    assert len(context) > 0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    self.merge_assembly(assembly)
    spaic.global_assembly_init_count -= 1

def select_assembly( assemblies, name=None):
    context = spaic.global_assembly_context_list
    assert len(context) > 0, "use global assembly build functions outside of the 'with' context"
    assert isinstance(context[-1], Assembly), "the object of the context is not an Assembly"
    spaic.global_assembly_init_count += 1
    self = context[-1]
    if name is None:
        new_asb = self.select_assembly(assemblies, context)
    else:
        new_asb = self.select_assembly(assemblies, name)
    spaic.global_assembly_init_count -= 1
    return new_asb


# -*- coding: utf-8 -*-
"""
Created on 2020/8/6
@project: SPAIC
@filename: Backend
@author: Hong Chaofei
@contact: hongchf@gmail.com
@description:
定义网络仿真使用的backend，如 Pytorch, Tensorflow, CUDA, 达尔文芯片等，以及相应的微分方程求解方法比如 Euler, 2阶 Runge-Kutta等
"""
import sys
from abc import abstractmethod, ABC
from collections import OrderedDict
from typing import Dict
from spaic.Network.BaseModule import BaseModule, VariableAgent, Op
from spaic.Network.DelayQueue import DelayQueue
from spaic.IO.Initializer import BaseInitializer
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

backends = dict()


class Backend(ABC):
    '''
    Basic backend class. All specified backend backend should subclass it.
    The backend is a parameter for the build function and becomes an attribute of all objects defined
    in the frontend backend network in building process. These objects build their initial data
    and specified operations into the attributes of backend, according to _variables
    and _operations respectively. The data will update in each step according the computation graph.
    Args:
        dt (float, optional): the length of a backend timestep, in millisecond.
    Attributes:
        device (str): the desired device of returned tensor. Its value can be 'cpu' or 'cuda'. If None, uses
            the current device for the default tensor type.
        builded (bool): whether the object defined in the frontend backend network has been builded.
        time (float): current backend time, in millisecond.
        n_time_step (int): the num of current time step.

        _variables (OrderedDict): records all variables from the build function of frontend objects.
        _parameters_dict (OrderedDict): records the variables to be trained.
        _InitVariables_dict (OrderedDict): reserves a copy of the initialization variables for initialization.
        _graph_var_dicts (dict): has following format: {'variables_dict': self._variables, 'temp_dict': dict(), 'update_dict': dict(), 'reduce_dict': dict()},
            recording the intermediate value of variables in computation progress.

        basic_operate (dict): dictionary of basic operators, mapping from operator names using in frontend to
            the funtion objects implemented in backend.
        _operations (list): records all basic operations from the build function of frontend objects, each of
            which has following format: [ret_var_name: str, operation_name, input_var_name1: str, input_var_name2 :str, ...].
        _graph_operations (list): redefine each basic operation, that is, add the corresponding keyword in the _graph_var_dicts to each variable,
            which has following format: [(dict_type, ret_var_name), operation_name, [(dict_type1, input_var_name1),(dict_type2, input_var_name2),...]].
        _standalone_operations (list): records all standalone operations from the build function of frontend objects,
            each of which has following format: (ret_var_name: str, function, input_var_names: list).
        _initial_operations (list): records all initial operations from the build function of frontend objects, each of
            which has following format: (ret_var_name: str, function, input_var_names: list).

        _monitors (list): records all monitors defined in fronted network through build function of Monitor object.

    Methods:
        build_graph: build a computation graph before performing the calculation.
        graph_update_step: update value of _graph_var_dicts.
        initial_step: initialize network variables.
        update_step: update the return variables of standalone operations and basic operations and current backend time.
        r_update_step: update the return variables of basic operations without using graph_update_step().
        add_variable: add variables from front objects to _variables of Backend.
        add_backend_variable: add variables according to the specified backend.
        add_operation: add basic operations from front objects to _operations of Backend.
        register_standalone: add standalone operations from front objects to _standalone_operations of Backend.
        register_initial: add initial operations from front objects to _initial_operations of Backend.
    '''
    basic_operate = dict()
    param_init_operate = dict()  # -> param_init_operate

    backend_name = 'None'
    def __init__(self, dt=0.1):
        super(Backend, self).__init__()
        self.device = None
        self.runtime = None
        self.builded = False
        self.partition = None
        self.dt = dt  # the length of a backend timestep
        self.time = 0.0  # current backend time
        self.n_time_step = 0  # the num of current time step
        self._batch_size = 1

        self._variables = dict()  # build from orderedDict to Tuple
        self._update_dict = dict()
        self._reduce_dict = dict()
        self._temp_dict = dict()
        self._parameters_dict = dict()
        self._clamp_parameter_dict = dict()
        self._delay_dict = dict()  # store conduction delays
        self._SparseVariables_dict = dict()
        self._InitVariables_dict = dict()
        self._variable_agent_dict: Dict[VariableAgent] = dict()
        self._temp_agent_dict: Dict[VariableAgent] = dict()

        self._operations = list()
        self._standalone_operations = list()
        self._initial_operations = list()

        self._monitors = list()  # TODO: need to add to update
        self._stored_states = dict()  # TODO: store network self._variables in the dict
        self.full_enable_grad = False
        self.forward_build = False #标识连接间有没有一步delay

        self.basic_operate['threshold'] = self.threshold
        self.basic_operate['reset'] = self.reset
        self.basic_operate['var_linear'] = self.var_linear
        self.basic_operate['mat_linear'] = self.mat_linear
        self.basic_operate['mat_mult_weight'] = self.mat_mult_weight
        self.basic_operate['mat_mult_weight_complex'] = self.mat_mult_weight_complex
        self.basic_operate['mat_mult_weight_2complex'] = self.mat_mult_weight_2complex
        self.basic_operate['mat_mult_pre'] = self.mat_mult_pre
        self.basic_operate['mat_mult'] = self.mat_mult
        self.basic_operate['bmm'] = self.bmm
        self.basic_operate['ger'] = self.ger
        self.basic_operate['sparse_mat_mult_weight'] = self.sparse_mat_mult_weight
        self.basic_operate['var_mult'] = self.var_mult
        self.basic_operate['add'] = self.add
        self.basic_operate['minus'] = self.minus
        self.basic_operate['div'] = self.div
        self.basic_operate['cat'] = self.cat
        self.basic_operate['stack'] = self.stack
        self.basic_operate['permute'] = self.permute
        self.basic_operate['view'] = self.view
        self.basic_operate['assign'] = self.assign
        self.basic_operate['unsqueeze'] = self.unsqueeze

        self.basic_operate['reduce_sum'] = self.reduce_sum
        self.basic_operate['conv_2d'] = self.conv_2d
        self.basic_operate['conv_2d_complex'] = self.conv_2d_complex
        self.basic_operate['relu'] = self.relu

        self.basic_operate['sin'] = self.sin
        self.basic_operate['cos'] = self.cos
        self.basic_operate['tan'] = self.tan
        self.basic_operate['log'] = self.log
        self.basic_operate['log2'] = self.log2
        self.basic_operate['log10'] = self.log10

        self.basic_operate['upsample'] = self.upsample
        self.basic_operate['conv_max_pool2d'] = self.conv_max_pool2d
        self.basic_operate['conv_avg_pool2d'] = self.conv_avg_pool2d
        self.basic_operate['conv_add_bias'] = self.conv_add_bias
        self.basic_operate['max_pool2d'] = self.max_pool2d
        self.basic_operate['post_max_pool2d_complex'] = self.post_max_pool2d_complex
        self.basic_operate['avg_pool2d'] = self.avg_pool2d
        self.basic_operate['batchnorm2d'] = self.batchnorm2d
        self.basic_operate['dropout'] = self.dropout
        self.basic_operate['reshape_mat_mult'] = self.reshape_mat_mult
        self.basic_operate['exp'] = self.exp
        self.basic_operate['mult_sum_weight'] = self.mult_sum_weight
        self.basic_operate['im2col_indices'] = self.im2col_indices
        self.basic_operate['conv2d_flatten'] = self.conv2d_flatten
        self.basic_operate['feature_map_flatten'] = self.feature_map_flatten
        self.basic_operate['weight_norm'] = self.weight_norm
        self.basic_operate['to'] = self.to

        self.param_init_operate['uniform'] = self.uniform
        self.param_init_operate['normal'] = self.normal
        self.param_init_operate['xavier_uniform'] = self.xavier_uniform
        self.param_init_operate['xavier_noraml'] = self.xavier_normal
        self.param_init_operate['kaiming_uniform'] = self.kaiming_uniform
        self.param_init_operate['kaiming_normal'] = self.kaiming_normal
        self.param_init_operate['constant'] = self.constant
        self.param_init_operate['sparse'] = self.sparse

        # self._graph_var_dicts = {'variables_dict': self._variables, 'temp_dict': dict(), 'update_dict': dict(),
        #                          'reduce_dict': dict()}

        self._graph_operations = list()
        self._push_operations = list()
        self._fetch_operations = list()

    def set_batch_size(self, batch_size):
        self._batch_size = batch_size

    def get_batch_size(self):
        return self._batch_size

    def set_runtime(self, runtime):
        self.runtime = runtime

    def register_basic_op(self, name, op_function):
        """
        register the function as basic operation to backend's _basic_operation dict
        Args:
            name:
            op_function:

        Returns:

        """
        name = name.lower()
        assert callable(op_function)
        if name in self.basic_operate:
            raise ValueError(('A model with the name "%s" has already been registered') % name)

        self.basic_operate[name] = op_function


    def build_graph(self):
        '''
        Build a computation graph before performing the calculation.
        Note that only the basic operations are redefiend into the _graph_operations list. The format of _graph_operations is as follows:
        [(dict_type, ret_var_name), operation_name, [(dict_type1, input_var_name1),(dict_type2, input_var_name2),...]].
        Traverse all basic operations and add the corresponding keyword in the _graph_var_dicts as dict_type to each variable in basic operation.
        '''

        variables_index = {k: i for i, k in enumerate(self._variables.keys())}

        self.initial_step()

        operation_type = 'update_dict or temp_dict or reduce_dict'
        # traverse basic operations

        fetch_operations = []
        push_operations = []
        graph_operations = []
        # self._graph_operations = list()
        # self._push_operations = list()
        # self._fetch_operations = list()

        for op in self._operations:
            if len(op.output) == 0 and len(op.input) == 0:
                # functions with no input and output will not push into the computation graph
                raise ValueError(" Operation lacks both input and output can't be build")
            elif len(op.output) == 0:
                fetch_operations.append(op)
            elif len(op.input) == 0:
                push_operations.append(op)
            else:
                graph_operations.append(op)

        ################################
        ##  for push_operation build  ##
        ################################
        update_dict = dict()
        reduce_dict = dict()

        for ind, op in enumerate(push_operations):
            outputs = []
            agent_outputs = []
            # if the operation return one variable, then it is appended into a list, to accordant with multi-variable returns
            if len(op.output) == 1:
                outputs.append(op.func())
            else:
                outputs = op.func()
            # return variable is a list
            for ind, var_name in enumerate(op.output):
                if var_name in self._variables:
                    # when the same ret_var_name occurs more than once, op.output is added to the reduce_dict of _graph_var_dicts
                    if var_name in update_dict:
                        reduce_dict[var_name] = [update_dict[var_name], outputs[ind]]
                        agent_outputs.append(self._variable_agent_dict[var_name].new_labeled_agent('reduce_dict'))
                        # # add op.output into graph: reduce_dict
                        # self._graph_var_dicts['reduce_dict'][op.output] = []
                        # revise the first reduce operation
                        for gop in self._push_operations:
                            for var in gop.output:
                                assert isinstance(var, VariableAgent)
                                if var.var_name == var_name:
                                    var.dict_label = 'reduce_dict'
                                    break
                        del update_dict[var_name]
                    elif var_name in reduce_dict:
                        reduce_dict[var_name].append(outputs[ind])
                        agent_outputs.append(self._variable_agent_dict[var_name].new_labeled_agent('reduce_dict'))
                    else:
                        # In the push_operation, new data is directly pushed to update_dict, as
                        # there is no need to remain the last step variable value
                        update_dict[var_name] = outputs[ind]
                        agent_outputs.append(self._variable_agent_dict[var_name].new_labeled_agent('update_dict'))
                else:
                    raise ValueError("No state variable to get the input ")

            # add the operation to built graph
            op.output = agent_outputs
            for agent in agent_outputs:
                agent.set_funcs.append(op)
            self._push_operations.append(op)

        # for var_name in reduce_dict:
        #     # add the reduce_sum operation into the graph
        #     self._graph_operations.append(
        #         [[('update_dict', var_name)], self.reduce_sum_update, [('reduce_dict', var_name)]])

        #################################
        ##  for graph_operation build  ##
        #################################
        temp_dict = dict()
        self.owners = set()
        temp_reduce_sum_ops = []
        self._temp_agent_dict = dict()
        for ind, op in enumerate(graph_operations):
            inputs = []
            agent_inputs = []
            for var_name in op.input:
                # try:
                #     var_name in self._variables
                # except:
                #     a = 1

                if '[updated]' in var_name:
                    var_name = var_name.replace("[updated]", "")

                    if var_name in update_dict:
                        inputs.append(update_dict[var_name])
                        agent_inputs.append(self._variable_agent_dict[var_name].new_labeled_agent('update_dict'))
                    elif var_name in reduce_dict:
                        # if the reduce_dict[var_name] is frozen: do reduce_sum operation before this op, and put the value to update_dict
                        # if not use the old variables dict

                        frozen = True
                        op_len = len(graph_operations)
                        for search_id in range(ind + 1, op_len):
                            if var_name in graph_operations[search_id].output:
                                frozen = False
                        if frozen: # all reduce_dict append ops have been done before this call
                            value = self.reduce_sum(self.stack(reduce_dict[var_name]))
                            inputs.append(value)
                            agent_inputs.append(self._variable_agent_dict[var_name].new_labeled_agent('update_dict'))
                            temp_reduce_sum_ops.append((var_name, len(reduce_dict[var_name])))
                            # add the reduce_sum operation into the graph
                            self._graph_operations.append(
                                Op([self._variable_agent_dict[var_name].new_labeled_agent('update_dict')],
                                   self.reduce_sum_update,
                                   [self._variable_agent_dict[var_name].new_labeled_agent('reduce_dict')],
                                   func=self.reduce_sum_update, owner=self, requires_grad=True))
                        else:
                            assert var_name in self._variables
                            inputs.append(self._variables[var_name])
                            agent_inputs.append(self._variable_agent_dict[var_name].new_labeled_agent('variables_dict'))
                    elif var_name in self._variables:
                        inputs.append(self._variables[var_name])
                        agent_inputs.append(self._variable_agent_dict[var_name].new_labeled_agent('variables_dict'))

                    else:
                        raise ValueError(" No State Variable [%s] in the update_dict" % var_name)
                elif var_name in self._variables:
                    inputs.append(self._variables[var_name])
                    if var_name not in self._variable_agent_dict:
                        self._variable_agent_dict[var_name] = VariableAgent(self, var_name)
                    agent_inputs.append(self._variable_agent_dict[var_name].new_labeled_agent('variables_dict'))


                elif var_name in temp_dict:
                    inputs.append(temp_dict[var_name])
                    if var_name in self._temp_agent_dict:
                        agent_inputs.append(self._temp_agent_dict[var_name])
                    else:
                        raise ValueError("asking a temp variable that has not been declared")
                else:
                    raise ValueError(" No State Variable [%s] in the variable dict" % var_name)

            outputs = []
            agent_outputs = []
            if len(op.output) == 0:
                self.var_check(op.func, inputs)
                op.func(*inputs)
            else:
                self.var_check(op.func, inputs)
                self.remove_tensor(inputs)
                if len(op.output) == 1:
                    outputs.append(op.func(*inputs))

                else:
                    outputs = op.func(*inputs)
                for ind, var_name in enumerate(op.output):
                    if var_name in self._variables:
                        # when the same ret_var_name occurs more than once, op.output is added to the reduce_dict of _graph_var_dicts
                        if var_name in update_dict:
                            reduce_dict[var_name] = [update_dict[var_name], outputs[ind]]
                            agent_outputs.append(self._variable_agent_dict[var_name].new_labeled_agent('reduce_dict'))
                            # # add op.output into graph: reduce_dict
                            # revise the first reduce operation
                            InGop = True
                            for pop in self._push_operations:
                                tmp_label_outputs = pop.output
                                for var in pop.output:
                                    if var.var_name == var_name:
                                        var.dict_label = 'reduce_dict'
                                        InGop = False
                                        break
                                else:
                                    continue
                                break
                            for gop in self._graph_operations:
                                # for all operations need the updated reduce_dict variable as input before reduce_sum are converted to use old variables_dict variables
                                for tmp_input in gop.input:
                                    if tmp_input.dict_label =='update_dict' and tmp_input.var_name == var_name:
                                        tmp_input.dict_label = 'variables_dict'
                                if InGop:
                                    for gop in self._graph_operations:
                                        for output in gop.output:
                                            if output.var_name == var_name:
                                                output.dict_label = 'reduce_dict'
                                                break
                            del update_dict[var_name]
                        elif var_name in reduce_dict:
                            reduce_dict[var_name].append(outputs[ind])
                            agent_outputs.append(self._variable_agent_dict[var_name].new_labeled_agent('reduce_dict'))
                        else:
                            update_dict[var_name] = outputs[ind]
                            agent_outputs.append(self._variable_agent_dict[var_name].new_labeled_agent('update_dict'))
                    else:
                        temp_dict[var_name] = outputs[ind]
                        if var_name in self._temp_agent_dict:
                            # raise ValueError("setting value to temp variable multiple times")
                            pass
                            # Whether allow duplication of name
                        else:
                            agent_outputs.append(VariableAgent(self, var_name, dict_label='temp_dict'))
                            self._temp_agent_dict[var_name] = agent_outputs[-1]


            # add the operation to built graph
            op.input = agent_inputs
            op.output = agent_outputs
            for agent in agent_inputs:
                agent.get_funcs.append(op)
            for agent in agent_outputs:
                agent.set_funcs.append(op)
            self._graph_operations.append(op)

        for reduce_op in temp_reduce_sum_ops:
            reduce_len = len(reduce_dict[reduce_op[0]])
            if reduce_len != reduce_op[1]:
                raise ValueError(
                    "Can't use [updated] tag for variable: %s, as it is a reduce_dict variable which is have updating conflict" %
                    reduce_op[0])
            else:
                del reduce_dict[reduce_op[0]]
        # for reduced variables that not used within [update]
        for var_name in reduce_dict:
            # add the reduce_sum operation into the graph
            self._graph_operations.append(
                Op([self._variable_agent_dict[var_name].new_labeled_agent('update_dict')],
                   self.reduce_sum_update, [self._variable_agent_dict[var_name].new_labeled_agent('reduce_dict')],
                   func=self.reduce_sum_update, owner=self, requires_grad=True))

        #################################
        ##  for fetch_operation build  ##
        #################################
        for ind, op in enumerate(fetch_operations):
            inputs = []
            agent_inputs = []
            for var_name in op.input:
                if '[updated]' in var_name:
                    # there is no need to have updated tag, as all variables computed in graph_operation have benn updated
                    var_name = var_name.replace("[updated]", "")
                if var_name in self._variables:
                    inputs.append(self._variables[var_name])
                    agent_inputs.append(self._variable_agent_dict[var_name].new_labeled_agent('variables_dict'))
                # elif var_name in temp_dict:
                #     inputs.append(temp_dict[var_name])
                #     label_inputs.append(('temp_dict', var_name))
                else:
                    raise ValueError(" No State Variable [%s] in the update_dict" % var_name)

            self.var_check(op.func, inputs)
            op.func(*inputs)

            op.input = agent_inputs
            # add the operation to built graph
            self._fetch_operations.append(op)

        # self._variables.update(update_dict)
        # for ii in range(len(self._graph_operations)):
        #     self._graph_operations[ii] = tuple(self._graph_operations[ii])
        self._graph_operations = tuple(self._graph_operations)

        self.set_func_grad()

        self.groups, self.isolate = self.get_dependency()
        self.get_place()

    def get_dependency(self):
        if self.partition:
            g = nx.DiGraph()
            for ind,op in enumerate(self._graph_operations):
                op.set_identifier(ind)
            g.add_nodes_from([op._identifier for op in self._graph_operations])
            for ind,op in enumerate(self._graph_operations):
                for output in op.output:
                    for get_op in output.get_funcs:
                        if op._identifier < get_op._identifier:
                            g.add_edge(op._identifier, get_op._identifier)

            plt.figure(figsize=(100,100))
            nx.draw_networkx(g)
            # plt.savefig('test.jpg')
            
            isolate = []
            leaf = []
            for i in range(len(self._graph_operations)):
                if g.degree(i) == 0:
                    isolate.append(i)
                else:
                    if g.out_degree(i) == 0:
                        leaf.append(i)

            g = g.to_undirected()
            groups = set()
            for i in leaf:
                group = frozenset(nx.dfs_tree(g,i))
                groups.add(group)
        else:
            groups = [self._graph_operations]
            isolate = []
        return groups, isolate

    def set_func_grad(self):
        if self.full_enable_grad:
            for op in self._operations:
                op.requires_grad = True
            for op in self._initial_operations:
                op.requires_grad = True

        for op in self._operations:
            if op.requires_grad:
                op.func = self.to_grad_func(op.func)
            else:
                op.func = self.to_nograd_func(op.func)

        for op in self._initial_operations:
            if op.requires_grad:
                op.func = self.to_grad_func(op.func)
            else:
                op.func = self.to_nograd_func(op.func)

    def var_check(self, op, *args):
        '''
        For specified operation, check the type or the shape of input variables.
        '''
        if op == 'mat_mult':
            if args[0][0].shape[1] != args[0][1].shape[0]:
                raise ValueError("%s and %s do not match" % (args[0].shape, args[1].shape))
        pass

    def graph_update_step_r(self):

        for op in self._graph_operations:
            inputs = []
            for var in op.input:
                inputs.append(self._graph_var_dicts[var[0]][var[1]])

            if op.output[0] is None:
                op.func(*inputs)
            elif op.output[0] == 'reduce_dict':
                self._graph_var_dicts['reduce_dict'][op.output[1]].append(op.func(*inputs))
            else:
                self._graph_var_dicts[op.output[0]][op.output[1]] = op.func(*inputs)

            # if '[updated]' in op.output[1]:
            #     op_name = op.output[1].strip('[updated]')
            #     if op_name in self._graph_var_dicts['update_dict'] and op_name in self._graph_var_dicts['variables_dict']:
            #         self._graph_var_dicts['update_dict'][op_name] = self._graph_var_dicts['temp_dict'][op.output[1]]  # 更新返回名中带[updated]的变量的值

        return  # tuple(self._graph_var_dicts['variables_dict'].values())

    def get_place(self):
        if self.partition:
            for ind,g in enumerate(self.groups):
                c = []
                g = [self._graph_operations[x] for x in g]
                for op in g:
                    for input in op.input:
                        if input._var_name.split(':')[-1] == '{weight}':
                            c.append(self._variables[input._var_name].device)
                place = max(set(c),key = c.count)
                for op in g:
                    op.place = place
            g = [self._graph_operations[x] for x in self.isolate]
            for ind,op in enumerate(g):
                for input in op.input:
                    if input._var_name.split(':')[-1] == '{weight}':
                        op.place = self._variables[input._var_name].device
                        break
                    else:
                        op.place = self.device[ind % self.device_count]
        else:
            for ind,op in enumerate(self._graph_operations):
                for input in op.input:
                    if input._var_name.split(':')[-1] == '{weight}':
                        op.place = self._variables[input._var_name].device
                        break
                    else:
                        op.place = self.device[ind % self.device_count]

    def graph_update_step(self):
        for op in self._graph_operations:
            # for inputs
            inputs = []
            for var in op.input:
                inputs.append(var.value)
            # compute the operation
            result = op.func(*inputs)
            if len(op.output) == 1: result = [result]
            # assign the result variables
            for ind, var in enumerate(op.output):
                var.value = result[ind]

    def push_update_step(self):
        for op in self._push_operations:
            result = op.func()
            if len(op.output) == 1: result = [result]
            for ind, var in enumerate(op.output):
                var.value = result[ind]

    def fetch_update_step(self):
        for op in self._fetch_operations:
            # for inputs
            inputs = []
            for var in op.input:
                inputs.append(var.value)
            op.func(*inputs)

    def initial_step(self):
        '''
        Initialize network variables.
        '''

        # Initialize the current backend time and the num of time step
        self.last_time = 0.0
        self.time = 0.0  # current backend time
        self.n_time_step = 0
        for key, value in self._variables.items():
            if '[stay]' in key:
                self._InitVariables_dict[key] = self._variables[key].detach()

        # Initialize untrainable variables
        self._variables.clear()

        # Initialize system spacial variables such as backend dt
        self._variables['[dt]'] = self.dt
        self._variables['[batch_size]'] = self._batch_size

        for key, value in self._InitVariables_dict.items():
            self._variables[key] = value

        # Initialize the trainable parameters
        for key, clamp_code in self._clamp_parameter_dict.items():
            clamp_code[0](*clamp_code[1])

        for key, value in self._parameters_dict.items():
            self._variables[key] = value

        for key, value in self._SparseVariables_dict.items():
            index_name = key + '_sparse_index'
            value_name = key + '_sparse_value'
            shape_name = key + '_sparse_shape'
            if index_name in self._variables.keys() and value_name in self._variables.keys():
                self._variables[key] = self.sparse_to_dense(index_name, value_name, shape_name)

        # Initialize the record of Monitor
        for monitor in self._monitors:
            monitor.init_record()

        # Traverse initial operations
        for op in self._initial_operations:
            inputs = []
            for var_name in op.input:
                if var_name in self._variables:
                    inputs.append(self._variables[var_name])
                else:
                    raise ValueError(" No State Variable [%s] in the variable dict" % var_name)
            if op.output is None:
                op.func(*inputs)
            else:
                result = op.func(*inputs)
                if len(op.output) == 1: result = [result]
                for ind, output in enumerate(op.output):
                    self._variables[output] = result[ind]

        # Change intial variable's batch_size
        for key in self._variables.keys():
            if hasattr(self._variables[key], 'shape'):
                shape = self._variables[key].shape
                if self._variables[key].ndim > 1 and shape[0] == 1 and (key not in self._parameters_dict):
                    expand_shape = -np.ones_like(shape, dtype=int)
                    expand_shape[0] = self._batch_size
                    self._variables[key] = self._variables[key].expand(tuple(expand_shape))

            # if '{O}' in key:
            #     o_shape = self._variables[key].shape
            #
            #     shape = []
            #     for s in o_shape:
            #         if s != 1:
            #             shape.append(s)
            #         else:
            #             shape.append(self._batch_size)
            #     self._variables[key] = torch.zeros(shape, dtype=torch.float32, device=self.device)

    def clear_step(self):
        '''

        Returns:

        '''

        self._operations = list()
        self._graph_operations = list()
        self._push_operations = list()
        self._fetch_operations = list()

    def initial_continue_step(self):
        '''
        Initialize network for continuous run.
        '''

        self.last_time = self.time

    def update_step(self):
        '''
        Update the return variables of standalone operations and basic operations and current backend time.
        Returns:
            tuple(self._variables.values())
        '''

        # push input data
        self.push_update_step()

        # static graph compuation
        if self.device_count > 1:
            self.graph_update_step_multigpu()
        else:
            self.graph_update_step()

        # Update time and state variables
        self.n_time_step += 1
        self.time = round(self.n_time_step * self.dt, 2)
        self._variables.update(self._update_dict)

        # fetch output data
        self.fetch_update_step()

        # Record Variables
        for monitor in self._monitors:
            monitor.update_step(self._variables)

        self._reduce_dict = dict()
        self._temp_dict = dict()
        self._update_dict = dict()

        return tuple(self._variables.values())

    def update_time_steps(self):
        while (self.runtime > self.time - self.last_time):
            self.update_step()

    def r_update_step(self):
        '''
        Update the return variables of basic operations without using graph_update_step().
        Returns:
            tuple(self._variables.values())
        '''

        reduce_dict = dict()
        self._graph_var_dicts['update_dict'].clear()
        self._graph_var_dicts['temp_dict'].clear()
        self._graph_var_dicts['reduce_dict'].clear()

        # Traverse standalone operations
        for op in self._standalone_operations:
            inputs = []
            for var_name in op.input:
                if 'pytorch' in backends:
                    inputs.append(self._variables[var_name])
                else:
                    inputs.append(self.to_numpy(self._variables[var_name]))

            if op.output is None:
                op.func(*inputs)
            else:
                if 'pytorch' in backends:
                    self._variables[op.output] = op.func(*inputs)
                else:
                    self._variables[op.output] = self.to_tensor(op.func(*inputs))

        # update one time_step
        for op in self._operations:
            if op.output in self._graph_var_dicts['variables_dict']:
                inputs = []
                for var_name in op.input:
                    if '[updated]' in var_name:
                        var_name = var_name.replace("[updated]", "")
                        if var_name in self._graph_var_dicts['update_dict']:
                            inputs.append(self._graph_var_dicts['update_dict'][var_name])
                        else:
                            raise ValueError(" No State Variable [%s] in the update_dict" % var_name)
                    elif var_name in self._graph_var_dicts['variables_dict']:
                        inputs.append(self._graph_var_dicts['variables_dict'][var_name])
                    elif var_name in self._graph_var_dicts['temp_dict']:
                        inputs.append(self._graph_var_dicts['temp_dict'][var_name])
                    else:
                        raise ValueError(" No State Variable [%s] in the variable dict" % var_name)

                if op.output in self._graph_var_dicts['update_dict']:
                    if op.output in self._graph_var_dicts['reduce_dict']:
                        self._graph_var_dicts['reduce_dict'][op.output].append(op.func(*inputs))
                    else:
                        self._graph_var_dicts['reduce_dict'][op.output] = [self._graph_var_dicts['update_dict'][op.output],
                                                                       op.func(*inputs)]
                else:
                    self._graph_var_dicts['update_dict'][op.output] = op.func(*inputs)
                    pass

            else:
                inputs = []
                for var_name in op.input:
                    if '[updated]' in var_name:
                        var_name = var_name.replace("[updated]", "")
                        if var_name in self._graph_var_dicts['update_dict']:
                            inputs.append(self._graph_var_dicts['update_dict'][var_name])
                        else:
                            raise ValueError(" No State Variable [%s] in the update_dict" % var_name)
                    elif var_name in self._graph_var_dicts['variables_dict']:
                        inputs.append(self._graph_var_dicts['variables_dict'][var_name])
                    elif var_name in self._graph_var_dicts['temp_dict']:
                        inputs.append(self._graph_var_dicts['temp_dict'][var_name])
                    else:
                        raise ValueError(" No State Variable [%s] in the variable dict" % var_name)
                self._graph_var_dicts['temp_dict'][op.output] = op.func(*inputs)

                if '[updated]' in op.output:
                    op_name = op.output.replace("[updated]", "")
                    if op_name in self._graph_var_dicts['update_dict']:
                        self._graph_var_dicts['update_dict'][op_name] = self._graph_var_dicts['temp_dict'][
                            op.output]  # update the variable in update_dict
                    else:
                        raise ValueError(" No State Variable [%s] in the update_dict" % var_name)

        # Update reduce_dict into update_dict
        for key, value in reduce_dict.items():
            value = self.stack(value)
            self._graph_var_dicts['update_dict'][key] = self.reduce_sum(value)
            self._graph_var_dicts['update_dict'][key] = []

        # update time
        self.n_time_step += 1
        self.time = round(self.n_time_step * self.dt, 2)

        self._graph_var_dicts['variables_dict'].update(self._graph_var_dicts['update_dict'])

        # Record Variables
        for monitor in self._monitors:
            monitor.update_step(self._graph_var_dicts)

        return tuple(self._variables.values())

    def reduce_sum_update(self, value):
        reduced = self.reduce_sum(self.stack(value))
        return reduced

    def get_varialble(self, name):

        if name in self._parameters_dict:
            return self._parameters_dict[name]
        elif name in self._variables:
            return self._variables[name]
        else:
            raise ValueError("not found variable:%s in the backend"%name)

    def set_variable_value(self, name, value, is_parameter):
        '''
        Set the backend value, in specific Backend
        Args:
            name:
            value:
            is_parameter:

        Returns:

        '''
        NotImplementedError()


    def add_variable(self, module: BaseModule, name: str, shape, value=None, is_parameter=False, is_sparse=False, init=None, init_param=None,
                     min=None, max=None, is_constant=False, prefer_device=None):
        '''
        Add variables from front objects to _variables of Backend and get copies to assign to _parameters_dict and _InitVariables_dict.
        Args:
            module (spaic.BaseModule) : the parent Module object the variable is belongs to
            name (str): the name of the added variable
            shape (list, int): the shape of the variable
            value (optional): the value of the variable
            is_parameter (bool, optional): whether the variable is trainable
            init (optional):
        '''
        if is_parameter:
            if isinstance(value, BaseInitializer):
                self._parameters_dict[name] = self.add_backend_variable(module, name, shape, None, grad=True, is_sparse=is_sparse,
                                                                        init=value.__class__.__name__, init_param=value.__dict__, prefer_device=prefer_device)
            else:
                self._parameters_dict[name] = self.add_backend_variable(module, name, shape, value, grad=True, is_sparse=is_sparse,
                                                                        init=init, init_param=init_param, prefer_device=prefer_device)
            # store clamp operations
            if min is not None and max is not None:
                self._clamp_parameter_dict[name] = (self.clamp_, [self._parameters_dict[name], min, max])
            elif min is not None:
                self._clamp_parameter_dict[name] = (self.clamp_min_, [self._parameters_dict[name], min])
            elif max is not None:
                self._clamp_parameter_dict[name] = (self.clamp_max_, [self._parameters_dict[name], max])


        # 稀疏矩阵weight非叶子节点，反传的时候更新的是weight中的value,但前向计算的时候用的是weight,所以对于稀疏矩阵要单独用个dict记录以便初始化
        elif is_sparse:
            self._SparseVariables_dict[name] = self.add_backend_variable(module, name, shape, value, grad=True,
                                                                         is_sparse=is_sparse, init=init,init_param=init_param, prefer_device=prefer_device)
        elif is_constant:
            self._InitVariables_dict[name] = value
        elif isinstance(value, BaseInitializer):
            self._InitVariables_dict[name] = self.add_backend_variable(module, name, shape, None, grad=False,
                                                                       is_sparse=is_sparse, init=value.__class__.__name__,
                                                                       init_param=value.__dict__, prefer_device=prefer_device)
        else:
            self._InitVariables_dict[name] = self.add_backend_variable(module, name, shape, value, grad=False,
                                                                       is_sparse=is_sparse, init=init,
                                                                       init_param=init_param, prefer_device=prefer_device)

        var_agent = VariableAgent(self, name, is_parameter)
        self._variable_agent_dict[name] = var_agent
        return var_agent

    def has_variable(self, name: str):
        if name.endswith('[updated]'):
            name = name.replace("[updated]", "")
        elif name.endswith('[stay]'):
            name = name.replace("[stay]", "")

        if name in self._variables:
            return True
        elif name in self._InitVariables_dict:
            return True
        elif name in self._parameters_dict:
            return True
        elif name in self._SparseVariables_dict:
            return True
        else:
            return False

    def add_delay(self, var_name, max_delay):
        max_len = int(max_delay / self.dt)
        if var_name in self._delay_dict:
            if self._delay_dict[var_name].max_len < max_len:
                self._delay_dict[var_name].max_len = max_len
        else:
            self._delay_dict[var_name] = DelayQueue(var_name, max_len, self)
            self.register_initial(Op(None, self._delay_dict[var_name].initial, [var_name, '[batch_size]'], owner=self, requires_grad=True))
            self.register_standalone(Op(None, self._delay_dict[var_name].push, [var_name, ], owner=self, requires_grad=True))
        return self._delay_dict[var_name]


    @abstractmethod
    def add_backend_variable(self, module: BaseModule, name, shape, value=None, grad=False, is_sparse=False, init=None, init_param=None):
        '''
        This method will be overwritten by different subclasses to add variables to _variables of specified backend.
        Args:
            module (spaic.BaseModule) : the parent Module object the variable is belongs to
            name (str): the name of the added variable
            shape (list, int): the shape of the variable
            value (optional): the value of the variable
            is_parameter (bool, optional): whether the variable is trainable
            init (optinal):
            grad (bool, optional): whether to use grad
        '''
        NotImplementedError()

    @abstractmethod
    def sparse_to_dense(self, index_name, value_name, shape_name):
        '''
        This method will be sparse matrix to dense matrix.
        Args:
            index_name (str)
            value_name (str)
            shape_name (str)
        '''
        NotImplementedError()

    def add_operation(self, op):
        '''
        Add basic operations from front objects to _operations of Backend.
        '''
        # op = Op()
        assert isinstance(op, Op)
        if isinstance(op.output, str):
            op.output = [op.output]
        elif op.output is None:
            op.output = []
        if isinstance(op.input, str):
            op.input = [op.input]
        elif op.input is None:
            op.input = []

        if op.func_name in self.basic_operate:
            op.func = self.basic_operate[op.func_name]
            if not isinstance(op.owner, Backend):
                assert isinstance(op.owner, BaseModule)
                op.owner._ops.append(op)
            op.operation_type = "_operations"
            self._operations.append(op)
        elif callable(op.func_name):
            self.register_standalone(op)
        else:
            raise ValueError("No operation %s in basic_operate" % op.func)


    def register_standalone(self, op):
        '''
        Add standalone operations from front objects to _standalone_operations of Backend.
        '''
        assert isinstance(op, Op)
        if isinstance(op.output, str):
            op.output = [op.output]
        elif op.output is None:
            op.output = []
        if isinstance(op.input, str):
            op.input = [op.input]
        elif op.input is None:
            op.input = []
        if not isinstance(op.owner, Backend):
            assert isinstance(op.owner, BaseModule)
            op.owner._ops.append(op)

        assert callable(op.func_name)
        op.func = op.func_name
        op.operation_type = "_operations"
        self._operations.append(op)


    def register_initial(self, op):
        '''
        Add initial operations from front objects to _initial_operations of Backend..
        op  = {output, func, input, owner, place, requires_grad}
        '''
        assert isinstance(op, Op)
        if isinstance(op.func, str):
            if op.func_name in self.basic_operate:
                op.func = self.basic_operate[op.func_name]
            else:
                raise ValueError("No operation %s in basic_operate" % op.func)
        else:
            assert callable(op.func_name)
            op.func = op.func_name
        if isinstance(op.input, str):
            op.input  = [op.input]
        elif op.input is None:
            op.input = []
        if isinstance(op.output, str):
            op.output = [op.output]
        elif op.output is None:
            op.output = []

        if not isinstance(op.owner, Backend):
            assert isinstance(op.owner, BaseModule)
            op.owner._ops.append(op)
        op.operation_type = "_initial_operations"
        self._initial_operations.append(op)

    def store(self, name='default'):
        '''
        Store backend_name and _variables into _stored_states dictionary.
        Args:
            name (str, optional): the name of network state.
        '''
        self._stored_states[name] = (self.backend_name, self._variables)

    def restore(self, name='default'):
        '''
        Restore network state from _stored_states dictionary.
        Args:
            name (str): the name of network state.
        '''
        if name not in self._stored_states:
            raise ValueError("No network state named: %s is stored" % name)
        else:
            stored_backend = self._stored_states[name][0]
            if stored_backend != self.backend_name:
                raise ValueError(
                    "The stored network is run by %s not %s" % (stored_backend, self.backend_name))
            else:
                self._variables = self._stored_states[name]

    def check_key(self, ckey, target_dict):
        cnetname = ckey[:ckey.find('<net>')]
        for key, value in target_dict.items():
            netname = key[:key.find('<net>')]
            break
        ckey = ckey.replace(cnetname, netname)
        if ckey in target_dict.keys():
            return ckey

        import warnings
        warnings.warn('Key error occurs, please check keys.')


        # result = [key for key in target_dict.keys() if key.endswith(variables[variables.find('<net>'):])]
        # if result:
        #     if len(result) > 1:
        #         import warnings
        #         warnings.warn('Given key matchs two variables in the backend dict, choose the first one as default')
        #     result = result[0]
        # return result

    @abstractmethod
    def to_nograd_func(self, func):
        '''
        Define the function as with no_grad
        Args:
            func:

        Returns:

        '''

    @abstractmethod
    def to_grad_func(self, func):
        '''
        Define the function as with enable_grad
        Args:
            func:

        Returns:

        '''

    # -------- basic backends operations -----
    @abstractmethod
    def threshold(self, v, v_th):
        '''
        Args:
            v: membrane voltage
            v_th: threshold
        Returns:
            v> v_th
        '''

    @abstractmethod
    def reset(self, v, o):
        '''

        Args:
            v: membrane voltage
            o: output spike

        Returns:
            if o>0, return 0
            else reutn v
        '''

    @abstractmethod
    def cat(self, x, dim=1):
        '''
        Joining data together along a dimension.
        Note that the total dimension of the data remains the same after cat.
        Args:
            x (list):
            dim (int): the dimension to cat.
        Returns:
            concat(x, dim)
        '''

    @abstractmethod
    def stack(self, x, dim=1):
        '''
        Add new dimension when stack data.
        Args:
            x (list):
            dim (int): the dimension to stack.
        Returns:
            stack(x, dim)
        '''

    @abstractmethod
    def permute(self, x, permute_dim):
        '''
        Parameters
        ----------
        x---> input
        permute_dim---> the dimension index of permute operation
        Returns
        -------
        '''

    @abstractmethod
    def view(self, x, view_dim):
        '''
        Parameters
        ----------
        x---> input
        view_dim---> the shape of view operation
        Returns
        -------
        '''

    def assign(self, x):
        '''
        Parameters
        ----------
        y---> target
        x---> input
        y = x
        Returns
        -------
        '''
        return x

    @abstractmethod
    def unsqueeze(self, x, dim):
        '''
        Parameters
        ----------
        x---> input
        dim---> the dim of unsqueeze operation
        Returns
        -------
        '''

    @abstractmethod
    def reduce_sum(self, x, *dim):
        '''
        Reduce the dimensions of the data
        Args:
            x (list):
            dim (tuple(int)): the dimension to reduce.
        Returns:
            sum(x, dim)
        '''

    @abstractmethod
    def index_select(self, x, indices, dim=1):
        '''
        Parameters
        ----------
        x
        indices
        Returns
        -------
        '''

    @abstractmethod
    def scatter(self, x, indices):
        '''
        Parameters
        ----------
        x
        indices
        Returns
        -------
        '''

    @abstractmethod
    def conv1d(self, x, kernel):
        '''
        Parameters
        ----------
        x
        kernel
        Returns
        -------
        '''
    @abstractmethod
    def conv_2d_complex(self, x, kernel, stride, padding, dilation, groups, beta, delay=None):
        '''

        Args:
            x:
            kernel:
            stride:
            padding:
            dilation:
            groups:
            beta:

        Returns:

        '''

    @abstractmethod
    def conv_trans2d(self, x, kernel, stride, padding, dilation, groups):
        '''
            transposed conv 2d
            Parameters
            ----------
            x
            kernel
            Returns
            -------
            '''
    @abstractmethod
    def conv_trans1d(self, x, kernel):
        '''
        transposed conv 1d
        Parameters
        ----------
        x
        kernel
        Returns
        -------
        '''

    @abstractmethod
    def im2col_indices(self, x, kh, kw, padding, stride):
        '''
        Parameters
        ----------
        x: 4D array  N, FH, FW, C_{in}
        kh: kernel_height
        kw: kernel_width
        stride:
        padding:
        Returns
        ----------
        '''

    @abstractmethod
    def conv2d_flatten(self, x):
        '''
        Parameters
        ----------
        x: 4D array (batch_size, out_channels, height, width)
        Returns
        3D array (batch_size, out_channels, height * width)
        ----------
        '''

    @abstractmethod
    def feature_map_flatten(self, x):
        '''
        For RSTDP and STDP learning rules which is  follwed with conv pre_layer
        Parameters
        ----------
        x: 4D array (batch_size, out_channels, height, width)
        Returns
        2D array (batch_size, out_channels * height * width)
        ----------
        '''

    @abstractmethod
    def add(self, x, y):
        '''
        Add the tensor y to the input x and returns a new result.
        Args:
            x (Tensor): input
            y (Tensor or Number): the second input
        Returns:
            x + y
        '''
        NotImplementedError()

    @abstractmethod
    def minus(self, x, y):
        '''
        The first input minus the second input
        Args:
            x (Tensor): input
            y (Tensor or Number): the second input
        Returns:
            x - y
        '''
        NotImplementedError()

    @abstractmethod
    def div(self, x, y):
        '''
        The first input div the second input
        Args:
            x (Tensor): input
            y (Tensor or Number): the second input

        Returns:
            x/y

        '''
        NotImplementedError()

    @abstractmethod
    def relu(self, x):
        '''
        Rectified Linear
        Args:
            x:

        Returns:
            x = x if x>0. else x = 0
        '''

    @abstractmethod
    def mat_mult_weight(self, A, X):
        '''
        Matrix product.
        Args:
            A (Tensor): the first input to be multiplied
            X (Tensor): the second input to be multiplied
        Returns:
            mat_mult_weight(A,X)
        '''
        NotImplementedError()

    @abstractmethod
    def mat_mult_weight_complex(self, A, X, beta, delay=None):
        NotImplementedError()
    @abstractmethod
    def mat_mult_weight_2complex(self, A, X, beta):
        NotImplementedError()

    @abstractmethod
    def mat_mult_pre(self, A, X):
        '''
        Matrix product.
        Args:
            A (Tensor): the first input to be multiplied
            X (Tensor): the second input to be multiplied
        Returns:
            mat_mult_pre(A,X)
        '''
        NotImplementedError()

    @abstractmethod
    def sigmoid(self, x):
        '''

        Args:
            x:

        Returns:

        '''

    @abstractmethod
    def upsample(self,x, scale):
        NotImplementedError()
    @abstractmethod
    def mat_mult(self, A, X):
        '''
        Matrix product.
        Args:
            A (Tensor): the first input to be multiplied
            X (Tensor): the second input to be multiplied
        Returns:
            mat_mult(A,X)
        '''
        NotImplementedError()

    @abstractmethod
    def reshape_mat_mult(self, A, X):
        '''
        Matrix product.
        Args:
            A (Tensor): the first input to be multiplied
            X (Tensor): the second input to be multiplied
        Returns:
        '''
        NotImplementedError()

    @abstractmethod
    def bmm(self, A, X):
        '''
        Performs a batch matrix-matrix product.
        Args:
            A (Tensor): the first input to be multiplied  [batch_size, n, m]
            X (Tensor): the second input to be multiplied  [batch_size, m, p]
        Returns:
            bmm(A,X)   [batch_size, n, p]
        '''
        NotImplementedError()

    @abstractmethod
    def sparse_mat_mult_weight(self, A, X):
        '''
        Sparse matrix product.
        Args:
            A (Tensor): the first input to be multiplied
            X (Tensor): the second input to be multiplied
        Returns:
            sparse_mat_mult_weight(A,X)
        '''
        NotImplementedError()

    @abstractmethod
    def var_mult(self, A, X):
        '''
        Args:
            A, X
        Returns:
            A * X
        '''
        NotImplementedError()

    @abstractmethod
    def mult_sum_weight(self, A, X):
        '''
         sum(A*X, dim=-2)
        Args:
            A:
            X:

        Returns:

        '''
        NotImplementedError()

    @abstractmethod
    def mat_linear(self, A, X, b):
        '''
        Args:
            A
            X
            b
        Returns:
            mat_mul(A,X)+b
        '''
        NotImplementedError()

    @abstractmethod
    def ger(self, A, X):
        '''
        Args:
            A
            X
        Returns:
            ger(A,X)
        '''
        NotImplementedError()

    @abstractmethod
    def var_linear(self, A, X, b):
        '''
        If A is matrix, then A and X should have the same shape, A*X is elemen-wise multiplication
        else  A should be a scalar value.
        Returns:
            A*X +b
        '''
        NotImplementedError()

    @abstractmethod
    def to_numpy(self, data):
        '''
        Args：
            data
        Returns:
            data.numpy()
        '''
        NotImplementedError()

    @abstractmethod
    def to_tensor(self, data):
        '''
        Args:
            data
        Returns:
            torch.tensor(data)
        '''
        NotImplementedError()

    @abstractmethod
    def clamp_(self, data, min, max):
        '''
            in-place clamp the data
        '''
        NotImplementedError()

    @abstractmethod
    def clamp_max_(self, data, max):
        '''
            in-place clamp the max of the data
        '''
        NotImplementedError()

    @abstractmethod
    def clamp_min_(self, data, min):
        '''
            in-place clamp the min of the data
        '''
        NotImplementedError()

    @abstractmethod
    def uniform(self, data, a=0.0, b=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a(float): the lower bound of the uniform distribution
            b(float): the upper bound of the uniform distribution
        Returns:
            torch.nn.init.uniform_(data, a=0.0, b=1.0)
        '''
        NotImplementedError()

    @abstractmethod
    def normal(self, data, mean=0.0, std=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            mean(float): the mean of the normal distribution
            std(float): the standard deviation of the normal distribution
        Returns:
            torch.nn.init.normal_(data, mean=0.0, std=1.0)
        '''
        NotImplementedError()

    @abstractmethod
    def xavier_normal(self, data, gain=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            gain: an optional scaling factor
        Returns:
            torch.nn.init.xavier_normal_(data, gain=1.0)
        '''
        NotImplementedError()

    @abstractmethod
    def xavier_uniform(self, data, gain=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            gain: an optional scaling factor
        Returns:
            torch.nn.init.xavier_uniform_(data, gain=1.0)
        '''
        NotImplementedError()

    @abstractmethod
    def kaiming_normal(self, data, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance of the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 'leaky_relu' (default).
        Returns:
            torch.nn.init.kaiming_normal_(data, a=0, mode='fan_in', nonlinearity='leaky_relu')
        '''
        NotImplementedError()

    @abstractmethod
    def kaiming_uniform(self, data, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance of the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 'leaky_relu' (default).
        Returns:
            torch.nn.init.kaiming_uniform_(data, a=0, mode='fan_in', nonlinearity='leaky_relu')
        '''
        NotImplementedError()

    @abstractmethod
    def constant(self, data, constant_value=0.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            constant_value(float): the value to fill the tensor with
        Returns:
            torch.nn.init.constant_(data, constant_value)
        '''
        NotImplementedError()

    @abstractmethod
    def to(self, data, device):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            constant_value(float): the value to fill the tensor with
        Returns:
            torch.nn.init.constant_(data, constant_value)
        '''
        NotImplementedError()

    @abstractmethod
    def weight_norm(self, weight, amp):
        '''

        w = g/||v|| * v
        Args:
            weight: v
            amp: g

        Returns: amp*weight/norm(weight)
        '''

    # @abstractmethod
    # def euler_update(self):
    #     pass
    #
    # @abstractmethod
    # def rk2_update(self):
    #     pass
    #
    # @abstractmethod
    # def reset(self, v, v_reset, u_reset, spike):
    #      '''
    #      voltage reset
    #
    #      Parameters
    #      ----------
    #      v
    #      v_reset
    #      u_reset
    #      spike
    #
    #      Returns
    #      -------
    #      v[spike] = v_reset
    #      v[spike] += u_reset
    #      '''
    #
    # @abstractmethod
    # def reset_u(self, u, u_reset, spike):
    #      '''
    #      recovery reset
    #
    #      Parameters
    #      ----------
    #      u
    #      _reset
    #      spike
    #
    #      Returns
    #      -------
    #      u[spike] = u+u_reset
    #      '''
    #      NotImplementedError()
    #
    # @abstractmethod
    # def next_stage(self, x):
    #     '''
    #
    #    Parameters
    #    ----------
    #    x: list
    #
    #    Returns
    #    -------
    #    x[index]
    #    '''
    #
    # @abstractmethod
    # def izh_v(self, v, u, psp):
    #     '''
    #
    #     Parameters
    #     ----------
    #     v: list
    #     u: list
    #     psp: list
    #
    #     Returns
    #     -------
    #     V=V+dt*(0.04*V^2+5*V+140-U+PSP)
    #     '''
    #     NotImplementedError()
    #
    # @abstractmethod
    # def izh_u(self, a, b, v, u):
    #     '''
    #
    #     Parameters
    #     ----------
    #     a: list
    #     b: list
    #     u: list
    #     v: list
    #
    #     Returns
    #     -------
    #     U=U+a*(b*V-U)
    #     '''
    #     NotImplementedError()


    def exp(self, x):
        '''
        Args:
            x(tensor): an n-dimensional torch.Tensor
        Returns:
           return exp(x)
        '''
        NotImplementedError()

    def sin(self, x):
        '''
        Args:
            x(tensor): an n-dimensional torch.Tensor
        Returns:
           return exp(x)
        '''
        NotImplementedError()

    def cos(self, x):
        '''
        Args:
            x(tensor): an n-dimensional torch.Tensor
        Returns:
           return exp(x)
        '''
        NotImplementedError()

    def tan(self, x):
        '''
        Args:
            x(tensor): an n-dimensional torch.Tensor
        Returns:
           return exp(x)
        '''
        NotImplementedError()

    def log(self, x):
        '''
        Args:
            x(tensor): an n-dimensional torch.Tensor
        Returns:
           return exp(x)
        '''
        NotImplementedError()

    def log2(self, x):
        '''
        Args:
            x(tensor): an n-dimensional torch.Tensor
        Returns:
           return exp(x)
        '''
        NotImplementedError()

    def log10(self, x):
        '''
        Args:
            x(tensor): an n-dimensional torch.Tensor
        Returns:
           return exp(x)
        '''
        NotImplementedError()




# class Darwin_Backend(Backend):
#
#     def __init__(self):
#         super(Darwin_Backend, self).__init__()
#         pass



# -*- coding: utf-8 -*-
"""
Created on 2022/5/30
@project: SPAIC
@filename: FiniteDifferenceGradientApproximation
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
A numerical analysis tool for approximating derivatives of networks using Finite Difference Method
"""
import spaic
from spaic.Network.BaseModule import Op
from .Learner import Learner
from spaic import Backend
import torch

class FiniteDifference(Learner):
    
    def __init__(self, trainable=None, **kwargs):
        super(FiniteDifference, self).__init__(trainable)
        self.index = kwargs.get('index', None)
        self.epsilon = kwargs.get('epsilon', 0.001)
        self.target = kwargs.get('target', None)
        self.var_name = kwargs.get('var_name', None)
        assert isinstance(self.index, list)
        assert isinstance(self.index[0], list)
        assert len(trainable) == 1 and isinstance(trainable[0], spaic.Connection)
        assert isinstance(self.var_name, str)
        self.record_values = []



    def weight_change(self, weight):
        new_weights_pos = []
        new_weights_neg = []
        for ind in self.index:
            pos_weight = weight.clone()
            neg_weight = weight.clone()
            pos_weight[ind[0], ind[1]] += self.epsilon
            neg_weight[ind[0], ind[1]] -= self.epsilon
            new_weights_pos.append(pos_weight)
            new_weights_neg.append(neg_weight)

        new_weights_neg = torch.stack(new_weights_neg, dim=0).unsqueeze(1)
        new_weights_pos = torch.stack(new_weights_pos, dim=0).unsqueeze(1)
        new_weights = torch.stack([new_weights_pos, new_weights_neg], dim=0)
        return new_weights

    def record_target(self, value):
        self.record_values.append(value)

        pass

    def build(self, backend):
        assert isinstance(self.target, spaic.BaseModule)
        assert isinstance(backend, spaic.Backend)
        self._backend = backend
        self.var_name = self.target.get_full_name(self.var_name)
        self.op_to_backend(None, self.record_target, self.var_name)











